import numpy as np
from numpy import matlib
from skimage.draw import polygon
from scipy.ndimage.morphology import binary_erosion
from scipy import sparse

def get_search_region(bbox, frame, ratio):
    x0, y0, w, h = bbox
    ih, iw, _ = frame.shape
    
    ww, hh = ratio*w, ratio*h
    
    # expand bbox by ratio
    x1 = np.min([iw-1, x0 + w/2 + ww/2])
    y1 = np.min([ih-1, y0 + h/2 + hh/2])
    x0 = np.max([0, x0 + w/2 - ww/2])
    y0 = np.max([0, y0 + h/2 - hh/2])
    
    x0, y0, x1, y1 = np.rint(np.array([x0, y0, x1, y1])).astype('int')
    
    return x0, y0, x1, y1, ww, hh

def coords2bb(x, y):
    # dim = 0 -> access pair of pts
    # dim = 1 -> access point
    # dim = 2 -> access x (0) or y (1) value
    return np.array([[[x[0], y[0]], [x[1], y[1]]],
                     [[x[1], y[1]], [x[2], y[2]]],
                     [[x[2], y[2]], [x[3], y[3]]],
                     [[x[3], y[3]], [x[0], y[0]]]], dtype='int')

def create_scribble_mask(x, y, frame, display=False, pct_area_shrink=0.1, pct_area_grow=0.1, winsz=3):
    # centroid
    xc, yc = x.mean(), y.mean()

    # expand bbox by 'pct_area_grow' area
    r = np.sqrt(1 + pct_area_grow)
    xe = r*(x - xc) + xc
    ye = r*(y - yc) + yc
    
    xe = np.maximum(np.minimum(xe, frame.shape[1]-1-winsz), winsz)
    ye = np.maximum(np.minimum(ye, frame.shape[0]-1-winsz), winsz)

    # shrink bbox by 'pct_area_shrink' area
    r = np.sqrt(1 - pct_area_shrink)
    xs = r*(x - xc) + xc
    ys = r*(y - yc) + yc

    xs = np.maximum(np.minimum(xs, frame.shape[1]-1-winsz), winsz)
    ys = np.maximum(np.minimum(ys, frame.shape[0]-1-winsz), winsz)
    
    # create scribble mask
    scribble_mask = np.zeros_like(frame[..., 0], dtype='int')
    expanded_mask = np.zeros_like(scribble_mask, dtype='bool')

    re, ce = polygon(xe, ye) # pixels for expanded polygon
    rs, cs = polygon(xs, ys) # shrank polygon

    expanded_mask[ce, re] = True

    scribble_mask.flat[~expanded_mask.ravel()] = -1
    scribble_mask[cs, rs] = 1

    return scribble_mask
    
def getC(mask, c):
    scribble_mask = mask != 0
    mask_area = np.prod(mask.shape)
    
    C = c * sparse.diags(scribble_mask.ravel(order='F'), offsets=0, 
                         shape=(mask_area, mask_area), dtype='float').tocsr()
    return C

def getAlpha_star(mask):
    alpha_star = np.zeros_like(mask, dtype='float')
    alpha_star.flat[mask.ravel() > 0] = 1
    alpha_star.flat[mask.ravel() < 0] = -1
    
    return alpha_star

def getLap(imdata, winsz, mask, _lambda, quiet=True):
    # convert to square window if only supplied side
    if type(winsz) == int:
        winsz = np.array([winsz, winsz])
        
    # normalize image data
    imdata = imdata.astype('float') / 255
    
    ih, iw, d = imdata.shape
    pixInds = np.reshape(np.arange(ih*iw), (ih, iw), order="F") # correct

    numPixInWindow = np.prod(winsz) # correct
    numPixInWindowSQ = numPixInWindow**2 # correct
    halfwinsz= ((winsz-1)/2).astype('int') # correct
    
    # erode scribble mask by window size
    scribble_mask = mask != 0 # correct
    
    struct_ele = np.ones((max(winsz), max(winsz))) # correct
    scribble_mask = binary_erosion(scribble_mask, struct_ele) # correct
    
    # storage for lap values
    numPix4Training = np.sum(1-scribble_mask[halfwinsz[0]+1:-halfwinsz[0], 
                                             halfwinsz[1]+1:-halfwinsz[1]])  # correct
    numNonzeroValue = numPix4Training * numPixInWindow**2  # correct
    
    row_inds = np.zeros(numNonzeroValue) # correct
    col_inds = np.zeros(numNonzeroValue) # correct
    vals = np.zeros(numNonzeroValue, dtype='float64') # correct
    
    # repeat on each legal pixel
    offset = 0
    
    winData = np.zeros((winsz[0], winsz[1], d))
    win_resized = np.zeros((numPixInWindow, d))
    
    if not quiet:
        pbar = tqdm(total=numNonzeroValue, leave=False)
    for i in range(halfwinsz[0], ih-halfwinsz[0]):
        for j in range(halfwinsz[1], iw-halfwinsz[1]):

            if scribble_mask[i, j]:
                continue

            # calclate indicies of pixels to place in the laplacian
            win_inds = pixInds[i-halfwinsz[0]:i+halfwinsz[0]+1,
                               j-halfwinsz[1]:j+halfwinsz[1]+1].ravel(order='F')
            row_inds[offset:offset+numPixInWindowSQ] = matlib.repmat(win_inds, 1, numPixInWindow)
            col_inds[offset:offset+numPixInWindowSQ] = matlib.repmat(win_inds, numPixInWindow, 1).ravel(order='F')

            # calculate the laplacian on a window
            winData[:] = imdata[i-halfwinsz[0]:i+halfwinsz[0]+1, j-halfwinsz[1]:j+halfwinsz[1]+1, :]
            win_resized[:] = np.reshape(winData, (numPixInWindow, d), order="F")
            vals[offset:offset+numPixInWindowSQ] = compLapCoeff(win_resized, numPixInWindow, d, _lambda).ravel(order='F')

            offset += numPixInWindowSQ
            if not quiet:
                pbar.update(numPixInWindowSQ)
    
    if not quiet:
        pbar.close()

    L = sparse.csr_matrix((vals, (row_inds, col_inds)), shape=(ih*iw, ih*iw))
    
    return L
    
def compLapCoeff(win, numPixInWindowSQ, d, _lambda):
    Xi = np.empty((numPixInWindowSQ, d+1))
    Xi[:, :-1] = win
    Xi[:, -1:] = 1

    I = np.eye(numPixInWindowSQ)
    I[-1, -1] = 0
    
    X = Xi @ Xi.T

    # need the transpose here as matrix is transpose of answer in matlab
    #F = (np.linalg.inv(X+(_lambda*I)) @ X).T
    F = np.linalg.solve(a=(X+(_lambda*I)).T, b=X.T).T
    
    I_F = np.eye(numPixInWindowSQ) - F
    
    return I_F.T @ I_F

def solveQuadOpt(L, C, alpha_star):
    lbda = 1e-6

    D = sparse.eye(L.shape[0], L.shape[1]).tocsr() # correct
    
    alpha = sparse.linalg.spsolve(L + C + (D*lbda), C @ alpha_star.ravel(order='F'))
    alpha = np.reshape(alpha, ((alpha_star.shape[0], alpha_star.shape[1])), order='F')
    
    # rescale alpha to ~ [0, 1]
    if np.min(alpha_star) == -1:
        alpha = alpha/2 + 0.5
    
    # clip to [0, 1]
    alpha = np.maximum(np.minimum(alpha, 1), 0)
    
    return alpha

def learningBasedMatting(frame, scribble_mask, winsz=3, c=800, _lambda=1e-07, quiet=True):
    L = getLap(frame, winsz, scribble_mask, _lambda, quiet=quiet)
    alpha = solveQuadOpt(L, getC(scribble_mask, c), getAlpha_star(scribble_mask))
    
    return alpha