import cython
import numpy as np
cimport numpy as np

# use the c sqrt function instead of numpy's
cdef extern from "math.h":
    double sqrt(double m)
    double pow(double m, double n)

    
############################################
######## ORIGINAL VERSION (using bc)
############################################

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def calc_cand_patch_likelihood(f, 
                               np.ndarray SS, 
                               np.ndarray CC, 
                               np.ndarray[np.int_t, ndim=3] GOPT_FP, 
                               np.intp_t dx, np.intp_t dy, np.int_t R):
    
    # number of feature channels (d is the channel no)
    cdef np.intp_t d_stop = <int> f.shape[2] - 1 # note we do 1 less to stop AFTER we hit it
    cdef np.intp_t d
    cdef np.ndarray[np.float_t, ndim=3] frame = f[:, :, ::1]
    
    # patch stuff
    cdef np.intp_t dx1 = dx + 1
    cdef np.intp_t dy1 = dy + 1
    cdef np.float_t patch_size = <float> (dx+dx1) * (dy+dy1)
    cdef np.float_t Rsq = R*R
    
    # size of samples/patches
    cdef np.intp_t n_samples = GOPT_FP.shape[0]
    cdef np.intp_t n_patches = GOPT_FP.shape[2]

    cdef np.ndarray[np.double_t, ndim=2] L = np.zeros((n_samples, n_patches), dtype=np.double)

    cdef np.intp_t i, j, ns, rx0, ry0, rx1, ry1, bsi, s, x, y
    cdef np.float_t dist, val, bdist, BC

    cdef np.ndarray[np.float_t, ndim=2] pmodel
    cdef np.ndarray[np.int_t, ndim=1] pcounts
    
    # preallocate histogram to be length of largest number of samples
    j = 0
    for x in range(len(CC)):
        j += CC[x].shape[0]
    cdef np.ndarray[np.int_t, ndim=1] hst = np.zeros(j, dtype=np.int)
    
    # for each patch
    for j in range(n_patches):
        pmodel  = SS[j] # patch model samples - size ()
        pcounts = CC[j] # patch sample counts
        
        ns = pcounts.shape[0]
        
        # check each candidate location for it - doing it this way avoids
        # accessing patch model and samples as often (only n_patches times)
        # and means we don't have to remake the temporary histogram array (hst)
        for i in range(n_samples):
            # patch location
            x = GOPT_FP[i, 0, j]
            y = GOPT_FP[i, 1, j]

            # patch bounds
            rx0 = x-dx ; ry0 = y-dy
            rx1 = x+dx1; ry1 = y+dy1

            # for each pixel
            for px in range(rx0, rx1):
                for py in range(ry0, ry1):
                    # best (matching) sample index reset to -1 (no match)
                    bsi = -1
                    bdist = Rsq

                    # for each sample
                    for s in range(ns):
                        
                        # calculate squared distance
                        dist = 0.0
                        d = 0
                        
                        while True:
                            val = pmodel[s, d] - frame[py, px, d]
                            dist += val * val
                            
                            # to avoid as many increments!
                            if d == d_stop:
                                break
                            
                            d += 1

                        # if it is < R**2 and the less than the current best distance..
                        if dist < Rsq:
                            if dist < bdist:
                                # it becomes the sample to match
                                bsi = s
                                bdist = dist

                    # if we got a match
                    if bsi > -1:
                        hst[bsi] += 1

            # calc Bhattacharyya coef candidate patch vs its model and
            # empty histogram for next candidate patch location
            BC = 0.0
            for s in range(ns):
                BC += sqrt(hst[s] * pcounts[s])
                hst[s] = 0 

            # normalise by size of patch and calculate Bhattacharyya distance
            L[i, j] = 1.0 - sqrt(1.0 - BC / patch_size)
    
    return L


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)    
def locally_optimise_cand_patches(np.ndarray[np.intp_t, ndim=1] inds,
                               f, 
                               np.ndarray SS, 
                               np.ndarray CC, 
                               np.ndarray[np.int_t, ndim=3] GOPT_FP,
                               np.ndarray[np.float_t, ndim=2] GOPT_FP_likelihood,
                               np.intp_t dx, np.intp_t dy, np.int_t R,
                               np.intp_t n_LWS,
                               np.ndarray[np.uint8_t, ndim=2, cast=True] local_opt_mask=None):
    
    # number of feature channels (d is the channel no)
    cdef ssize_t d_stop = <int> f.shape[2] - 1 # note we do 1 less to stop AFTER we hit it
    cdef ssize_t d
    cdef np.ndarray[np.float_t, ndim=3] frame = f[:, :, ::1]
    
    # patch stuff
    cdef np.intp_t dx1 = dx + 1
    cdef np.intp_t dy1 = dy + 1
    cdef np.float_t patch_size = <float> (dx+dx1) * (dy+dy1)
    cdef np.float_t Rsq = R*R
    
    # size of samples/patches
    cdef np.intp_t n_samples = inds.shape[0]
    cdef np.intp_t n_patches = GOPT_FP.shape[2]
    
    # local optimisation window size
    cdef np.intp_t nx0 = -(n_LWS // 2)
    cdef np.intp_t ny0 = -(n_LWS // 2)
    
    cdef np.intp_t nx1 = (n_LWS // 2) + 1
    cdef np.intp_t ny1 = (n_LWS // 2) + 1
    
    # pointers of stuff
    cdef np.intp_t j, ns, i, x, y, rx0, ry0, rx1, ry1, sample_index, s, bsi
    cdef np.intp_t rx0o, rx1o, ry0o, ry1o, xo, yo, best_x, best_y, py, px, d1, d2
    
    cdef np.float_t dist, val, bdist, BC, BD, best_location_likelihood

    cdef np.ndarray[np.float_t, ndim=2] pmodel
    cdef np.ndarray[np.int_t, ndim=1] pcounts
    
    # preallocate histogram to be length of largest number of samples
    j = 0
    for x in range(len(CC)):
        j += CC[x].shape[0]
    cdef np.ndarray[np.int_t, ndim=1] hst = np.zeros(j, dtype=np.int)
    
    # check if local_opt_mask is None -> locally optimise all patches
    if local_opt_mask is None:
        local_opt_mask = np.ones((n_samples, n_patches), dtype=np.bool)
    
    # for each patch
    for j in range(n_patches):
        pmodel  = SS[j] # patch model samples - size ()
        pcounts = CC[j] # patch sample counts
        
        ns = pcounts.shape[0]
        
        # for each candidate patch to be locally optimised
        for i in range(n_samples):
            # skip if we aren't locally optimising
            if not local_opt_mask[i, j]:
                continue
                
            sample_index = inds[i]
            
            # patch location
            x = GOPT_FP[sample_index, 0, j]
            y = GOPT_FP[sample_index, 1, j]

            # base patch bounds
            rx0 = x-dx ; ry0 = y-dy
            rx1 = x+dx1; ry1 = y+dy1
            
            # set negative likelihood to beat as best pos
            best_location_likelihood = 0
            best_x = 0
            best_y = 0
            
            # evaluate each possible move
            for xo in range(nx0, nx1):
                for yo in range(ny0, nx1):
                    # candidate local opt position
                    rx0o = rx0 + xo
                    rx1o = rx1 + xo
                    ry0o = ry0 + yo
                    ry1o = ry1 + yo
                    
                    # for each pixel
                    for py in range(ry0o, ry1o):
                        for px in range(rx0o, rx1o):
                            # best (matching) sample index reset to -1 (no match)
                            bsi = -1
                            bdist = Rsq

                            # for each sample
                            for s in range(ns):
                                # calculate squared distance
                                dist = 0.0
                                d = 0
                                while True:
                                    val = pmodel[s, d] - frame[py, px, d]
                                    dist += val * val
                                    
                                    if d == d_stop:
                                        break
                                        
                                    d += 1

                                # if it is < R**2 and the less than the current best distance..
                                if dist < Rsq:
                                    if dist < bdist:
                                        # it becomes the sample to match
                                        bsi = s
                                        bdist = dist

                            # if we got a match
                            if bsi > -1:
                                hst[bsi] += 1

                    # calc Bhattacharyya coef candidate patch vs its model and
                    # empty histogram for next candidate patch location
                    BC = 0.0
                    for s in range(ns):
                        BC += sqrt(hst[s] * pcounts[s])
                        hst[s] = 0 
                    BD = 1.0 - sqrt(1.0 - BC / patch_size)

                    # higher likelihood than the current best
                    if BD > best_location_likelihood:
                        best_x = xo
                        best_y = yo
                        best_location_likelihood = BD
                        
                    # equal likelihood
                    elif BD == best_location_likelihood:
                        # so decide based on distance from (0, 0)
                        d1 = xo*xo + yo*yo
                        d2 = best_x*best_x + best_y*best_y
                        if d1 < d2:
                            best_x = xo
                            best_y = yo
                            best_location_likelihood = BD
              
            # move patch to optimal location and save its likelihood
            GOPT_FP[sample_index, 0, j] += <int> best_x
            GOPT_FP[sample_index, 1, j] += <int> best_y
            GOPT_FP_likelihood[sample_index, j] = best_location_likelihood


############################################
######## SQUARED VERSION (squared bc)
############################################

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def calc_cand_patch_likelihood_squared(f, 
                               np.ndarray SS, 
                               np.ndarray CC, 
                               np.ndarray[np.int_t, ndim=3] GOPT_FP, 
                               np.intp_t dx, np.intp_t dy, np.int_t R):
    
    # number of feature channels (d is the channel no)
    cdef ssize_t d_stop = <int> f.shape[2] - 1 # note we do 1 less to stop AFTER we hit it
    cdef ssize_t d
    cdef np.ndarray[np.float_t, ndim=3] frame = f[:, :, ::1]
    
    # patch stuff
    cdef np.intp_t dx1 = dx + 1
    cdef np.intp_t dy1 = dy + 1
    cdef np.float_t patch_size = <float> (dx+dx1) * (dy+dy1)
    cdef np.float_t Rsq = R*R
    
    # size of samples/patches
    cdef np.intp_t n_samples = GOPT_FP.shape[0]
    cdef np.intp_t n_patches = GOPT_FP.shape[2]

    cdef np.ndarray[np.double_t, ndim=2] L = np.zeros((n_samples, n_patches), dtype=np.double)

    cdef np.intp_t i, j, ns, rx0, ry0, rx1, ry1, bsi, s, x, y
    cdef np.float_t dist, val, bdist, BC

    cdef np.ndarray[np.float_t, ndim=2] pmodel
    cdef np.ndarray[np.int_t, ndim=1] pcounts
    
    # preallocate histogram to be length of largest number of samples
    j = 0
    for x in range(len(CC)):
        j += CC[x].shape[0]
    cdef np.ndarray[np.int_t, ndim=1] hst = np.zeros(j, dtype=np.int)
    
    # for each patch
    for j in range(n_patches):
        pmodel  = SS[j] # patch model samples - size ()
        pcounts = CC[j] # patch sample counts
        
        ns = pcounts.shape[0]
        
        # check each candidate location for it - doing it this way avoids
        # accessing patch model and samples as often (only n_patches times)
        # and means we don't have to remake the temporary histogram array (hst)
        for i in range(n_samples):
            # patch location
            x = GOPT_FP[i, 0, j]
            y = GOPT_FP[i, 1, j]

            # patch bounds
            rx0 = x-dx ; ry0 = y-dy
            rx1 = x+dx1; ry1 = y+dy1

            # for each pixel
            for px in range(rx0, rx1):
                for py in range(ry0, ry1):
                    # best (matching) sample index reset to -1 (no match)
                    bsi = -1
                    bdist = Rsq

                    # for each sample
                    for s in range(ns):
                        
                        # calculate squared distance
                        dist = 0.0
                        d = 0
                        
                        while True:
                            val = pmodel[s, d] - frame[py, px, d]
                            dist += val * val
                            
                            # to avoid as many increments!
                            if d == d_stop:
                                break
                            
                            d += 1

                        # if it is < R**2 and the less than the current best distance..
                        if dist < Rsq:
                            if dist < bdist:
                                # it becomes the sample to match
                                bsi = s
                                bdist = dist

                    # if we got a match
                    if bsi > -1:
                        hst[bsi] += 1

            # calc Bhattacharyya coef candidate patch vs its model and
            # empty histogram for next candidate patch location
            BC = 0.0
            for s in range(ns):
                BC += sqrt(hst[s] * pcounts[s])
                hst[s] = 0 

            # normalise by size of patch and calculate Bhattacharyya distance
            L[i, j] = BC / patch_size
    
    return L


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)    
def locally_optimise_cand_patches_squared(np.ndarray[np.intp_t, ndim=1] inds,
                               f, 
                               np.ndarray SS, 
                               np.ndarray CC, 
                               np.ndarray[np.int_t, ndim=3] GOPT_FP,
                               np.ndarray[np.float_t, ndim=2] GOPT_FP_likelihood,
                               np.intp_t dx, np.intp_t dy, np.int_t R,
                               np.intp_t n_LWS,
                               np.ndarray[np.uint8_t, ndim=2, cast=True] local_opt_mask=None):
    
    # number of feature channels (d is the channel no)
    cdef ssize_t d_stop = <int> f.shape[2] - 1 # note we do 1 less to stop AFTER we hit it
    cdef ssize_t d
    cdef np.ndarray[np.float_t, ndim=3] frame = f[:, :, ::1]
    
    # patch stuff
    cdef np.intp_t dx1 = dx + 1
    cdef np.intp_t dy1 = dy + 1
    cdef np.float_t patch_size = <float> (dx+dx1) * (dy+dy1)
    cdef np.float_t Rsq = R*R
    
    # size of samples/patches
    cdef np.intp_t n_samples = inds.shape[0]
    cdef np.intp_t n_patches = GOPT_FP.shape[2]
    
    # local optimisation window size
    cdef np.intp_t nx0 = -(n_LWS // 2)
    cdef np.intp_t ny0 = -(n_LWS // 2)
    
    cdef np.intp_t nx1 = (n_LWS // 2) + 1
    cdef np.intp_t ny1 = (n_LWS // 2) + 1
    
    # pointers of stuff
    cdef np.intp_t j, ns, i, x, y, rx0, ry0, rx1, ry1, sample_index, s, bsi
    cdef np.intp_t rx0o, rx1o, ry0o, ry1o, xo, yo, best_x, best_y, py, px, d1, d2
    
    cdef np.float_t dist, val, bdist, BC, BD, best_location_likelihood

    cdef np.ndarray[np.float_t, ndim=2] pmodel
    cdef np.ndarray[np.int_t, ndim=1] pcounts
    
    # preallocate histogram to be length of largest number of samples
    j = 0
    for x in range(len(CC)):
        j += CC[x].shape[0]
    cdef np.ndarray[np.int_t, ndim=1] hst = np.zeros(j, dtype=np.int)
    
    # check if local_opt_mask is None -> locally optimise all patches
    if local_opt_mask is None:
        local_opt_mask = np.ones((n_samples, n_patches), dtype=np.bool)
    
    # for each patch
    for j in range(n_patches):
        pmodel  = SS[j] # patch model samples - size ()
        pcounts = CC[j] # patch sample counts
        
        ns = pcounts.shape[0]
        
        # for each candidate patch to be locally optimised
        for i in range(n_samples):
            # skip if we aren't locally optimising
            if not local_opt_mask[i, j]:
                continue
                
            sample_index = inds[i]
            
            # patch location
            x = GOPT_FP[sample_index, 0, j]
            y = GOPT_FP[sample_index, 1, j]

            # base patch bounds
            rx0 = x-dx ; ry0 = y-dy
            rx1 = x+dx1; ry1 = y+dy1
            
            # set negative likelihood to beat as best pos
            best_location_likelihood = 0
            best_x = 0
            best_y = 0
            
            # evaluate each possible move
            for xo in range(nx0, nx1):
                for yo in range(ny0, nx1):
                    # candidate local opt position
                    rx0o = rx0 + xo
                    rx1o = rx1 + xo
                    ry0o = ry0 + yo
                    ry1o = ry1 + yo
                    
                    # for each pixel
                    for py in range(ry0o, ry1o):
                        for px in range(rx0o, rx1o):
                            # best (matching) sample index reset to -1 (no match)
                            bsi = -1
                            bdist = Rsq

                            # for each sample
                            for s in range(ns):
                                # calculate squared distance
                                dist = 0.0
                                d = 0
                                while True:
                                    val = pmodel[s, d] - frame[py, px, d]
                                    dist += val * val
                                    
                                    if d == d_stop:
                                        break
                                        
                                    d += 1

                                # if it is < R**2 and the less than the current best distance..
                                if dist < Rsq:
                                    if dist < bdist:
                                        # it becomes the sample to match
                                        bsi = s
                                        bdist = dist

                            # if we got a match
                            if bsi > -1:
                                hst[bsi] += 1

                    # calc Bhattacharyya coef candidate patch vs its model and
                    # empty histogram for next candidate patch location
                    BC = 0.0
                    for s in range(ns):
                        BC += sqrt(hst[s] * pcounts[s])
                        hst[s] = 0 
                    BD = BC / patch_size

                    # higher likelihood than the current best
                    if BD > best_location_likelihood:
                        best_x = xo
                        best_y = yo
                        best_location_likelihood = BD
                        
                    # equal likelihood
                    elif BD == best_location_likelihood:
                        # so decide based on distance from (0, 0)
                        d1 = xo*xo + yo*yo
                        d2 = best_x*best_x + best_y*best_y
                        if d1 < d2:
                            best_x = xo
                            best_y = yo
                            best_location_likelihood = BD
              
            # move patch to optimal location and save its likelihood
            GOPT_FP[sample_index, 0, j] += <int> best_x
            GOPT_FP[sample_index, 1, j] += <int> best_y
            GOPT_FP_likelihood[sample_index, j] = best_location_likelihood
            
            
############################################
######## MINKOWSKI VERSION (any power on BD) - i.e. not square root
############################################


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def calc_cand_patch_likelihood_minkowski(f, 
                               np.ndarray SS, 
                               np.ndarray CC, 
                               np.ndarray[np.int_t, ndim=3] GOPT_FP, 
                               np.intp_t dx, np.intp_t dy, np.int_t R,
                               np.double_t power=1.0):
    
    # number of feature channels (d is the channel no)
    cdef ssize_t d_stop = <int> f.shape[2] - 1 # note we do 1 less to stop AFTER we hit it
    cdef ssize_t d
    cdef np.ndarray[np.float_t, ndim=3] frame = f[:, :, ::1]
    
    # patch stuff
    cdef np.intp_t dx1 = dx + 1
    cdef np.intp_t dy1 = dy + 1
    cdef np.double_t patch_size = <float> (dx+dx1) * (dy+dy1)
    cdef np.float_t Rsq = R*R
    
    # size of samples/patches
    cdef np.intp_t n_samples = GOPT_FP.shape[0]
    cdef np.intp_t n_patches = GOPT_FP.shape[2]

    cdef np.ndarray[np.double_t, ndim=2] L = np.zeros((n_samples, n_patches), dtype=np.double)

    cdef np.intp_t i, j, ns, rx0, ry0, rx1, ry1, bsi, s, x, y
    cdef np.float_t dist, val, bdist
    cdef np.double_t BC

    cdef np.ndarray[np.float_t, ndim=2] pmodel
    cdef np.ndarray[np.int_t, ndim=1] pcounts
    
    # preallocate histogram to be length of largest number of samples
    j = 0
    for x in range(len(CC)):
        j += CC[x].shape[0]
    cdef np.ndarray[np.int_t, ndim=1] hst = np.zeros(j, dtype=np.int)
    
    # for each patch
    for j in range(n_patches):
        pmodel  = SS[j] # patch model samples - size ()
        pcounts = CC[j] # patch sample counts
        
        ns = pcounts.shape[0]
        
        # check each candidate location for it - doing it this way avoids
        # accessing patch model and samples as often (only n_patches times)
        # and means we don't have to remake the temporary histogram array (hst)
        for i in range(n_samples):
            # patch location
            x = GOPT_FP[i, 0, j]
            y = GOPT_FP[i, 1, j]

            # patch bounds
            rx0 = x-dx ; ry0 = y-dy
            rx1 = x+dx1; ry1 = y+dy1

            # for each pixel
            for px in range(rx0, rx1):
                for py in range(ry0, ry1):
                    # best (matching) sample index reset to -1 (no match)
                    bsi = -1
                    bdist = Rsq

                    # for each sample
                    for s in range(ns):
                        
                        # calculate squared distance
                        dist = 0.0
                        d = 0
                        
                        while True:
                            val = pmodel[s, d] - frame[py, px, d]
                            dist += val * val
                            
                            # to avoid as many increments!
                            if d == d_stop:
                                break
                            
                            d += 1

                        # if it is < R**2 and the less than the current best distance..
                        if dist < Rsq:
                            if dist < bdist:
                                # it becomes the sample to match
                                bsi = s
                                bdist = dist

                    # if we got a match
                    if bsi > -1:
                        hst[bsi] += 1

            BC = 0.0
            for s in range(ns):
                BC += sqrt(hst[s] * pcounts[s])
                hst[s] = 0 
            
            # normalise by patch size so BC in [0, 1]
            # calculate Bhattacharyya distance (with minowski change)
            L[i, j] = 1.0 - pow(1.0 - BC / patch_size, power)
    
    return L
    
    
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)    
def locally_optimise_cand_patches_minkowski(np.ndarray[np.intp_t, ndim=1] inds,
                               f, 
                               np.ndarray SS, 
                               np.ndarray CC, 
                               np.ndarray[np.int_t, ndim=3] GOPT_FP,
                               np.ndarray[np.float_t, ndim=2] GOPT_FP_likelihood,
                               np.intp_t dx, np.intp_t dy, np.int_t R,
                               np.intp_t n_LWS,
                               np.ndarray[np.uint8_t, ndim=2, cast=True] local_opt_mask=None,
                               np.double_t power=1.0):
    
    # number of feature channels (d is the channel no)
    cdef ssize_t d_stop = <int> f.shape[2] - 1 # note we do 1 less to stop AFTER we hit it
    cdef ssize_t d
    cdef np.ndarray[np.float_t, ndim=3] frame = f[:, :, ::1]
    
    # patch stuff
    cdef np.intp_t dx1 = dx + 1
    cdef np.intp_t dy1 = dy + 1
    cdef np.float_t patch_size = <float> (dx+dx1) * (dy+dy1)
    cdef np.float_t Rsq = R*R
    
    # size of samples/patches
    cdef np.intp_t n_samples = inds.shape[0]
    cdef np.intp_t n_patches = GOPT_FP.shape[2]
    
    # local optimisation window size
    cdef np.intp_t nx0 = -(n_LWS // 2)
    cdef np.intp_t ny0 = -(n_LWS // 2)
    
    cdef np.intp_t nx1 = (n_LWS // 2) + 1
    cdef np.intp_t ny1 = (n_LWS // 2) + 1
    
    # pointers of stuff
    cdef np.intp_t j, ns, i, x, y, rx0, ry0, rx1, ry1, sample_index, s, bsi
    cdef np.intp_t rx0o, rx1o, ry0o, ry1o, xo, yo, best_x, best_y, py, px, d1, d2
    
    cdef np.float_t dist, val, bdist, BC, BD, best_location_likelihood

    cdef np.ndarray[np.float_t, ndim=2] pmodel
    cdef np.ndarray[np.int_t, ndim=1] pcounts
    
    # preallocate histogram to be length of largest number of samples
    j = 0
    for x in range(len(CC)):
        j += CC[x].shape[0]
    cdef np.ndarray[np.int_t, ndim=1] hst = np.zeros(j, dtype=np.int)
    
    # check if local_opt_mask is None -> locally optimise all patches
    if local_opt_mask is None:
        local_opt_mask = np.ones((n_samples, n_patches), dtype=np.bool)
    
    # for each patch
    for j in range(n_patches):
        pmodel  = SS[j] # patch model samples - size ()
        pcounts = CC[j] # patch sample counts
        
        ns = pcounts.shape[0]
        
        # for each candidate patch to be locally optimised
        for i in range(n_samples):
            # skip if we aren't locally optimising
            if not local_opt_mask[i, j]:
                continue
                
            sample_index = inds[i]
            
            # patch location
            x = GOPT_FP[sample_index, 0, j]
            y = GOPT_FP[sample_index, 1, j]

            # base patch bounds
            rx0 = x-dx ; ry0 = y-dy
            rx1 = x+dx1; ry1 = y+dy1
            
            # set negative likelihood to beat as best pos
            best_location_likelihood = 0
            best_x = 0
            best_y = 0
            
            # evaluate each possible move
            for xo in range(nx0, nx1):
                for yo in range(ny0, nx1):
                    # candidate local opt position
                    rx0o = rx0 + xo
                    rx1o = rx1 + xo
                    ry0o = ry0 + yo
                    ry1o = ry1 + yo
                    
                    # for each pixel
                    for py in range(ry0o, ry1o):
                        for px in range(rx0o, rx1o):
                            # best (matching) sample index reset to -1 (no match)
                            bsi = -1
                            bdist = Rsq

                            # for each sample
                            for s in range(ns):
                                # calculate squared distance
                                dist = 0.0
                                d = 0
                                while True:
                                    val = pmodel[s, d] - frame[py, px, d]
                                    dist += val * val
                                    
                                    if d == d_stop:
                                        break
                                        
                                    d += 1

                                # if it is < R**2 and the less than the current best distance..
                                if dist < Rsq:
                                    if dist < bdist:
                                        # it becomes the sample to match
                                        bsi = s
                                        bdist = dist

                            # if we got a match
                            if bsi > -1:
                                hst[bsi] += 1

                    # calc Bhattacharyya coef candidate patch vs its model and
                    # empty histogram for next candidate patch location
                    BC = 0.0
                    for s in range(ns):
                        BC += sqrt(hst[s] * pcounts[s])
                        hst[s] = 0 
                    BD = 1.0 - pow(1.0 - BC / patch_size, power)

                    # higher likelihood than the current best
                    if BD > best_location_likelihood:
                        best_x = xo
                        best_y = yo
                        best_location_likelihood = BD
                        
                    # equal likelihood
                    elif BD == best_location_likelihood:
                        # so decide based on distance from (0, 0)
                        d1 = xo*xo + yo*yo
                        d2 = best_x*best_x + best_y*best_y
                        if d1 < d2:
                            best_x = xo
                            best_y = yo
                            best_location_likelihood = BD
              
            # move patch to optimal location and save its likelihood
            GOPT_FP[sample_index, 0, j] += <int> best_x
            GOPT_FP[sample_index, 1, j] += <int> best_y
            GOPT_FP_likelihood[sample_index, j] = best_location_likelihood

            
############################################
######## MINKOWSKI VERSION (any power on BD) - WITH FLOATING POINT COUNTS
############################################


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def calc_cand_patch_likelihood_minkowski_floatcount(f, 
                               np.ndarray SS, 
                               np.ndarray CC, 
                               np.ndarray[np.int_t, ndim=3] GOPT_FP, 
                               np.intp_t dx, np.intp_t dy, np.int_t R,
                               np.double_t power=1.0):
    
    # number of feature channels (d is the channel no)
    cdef ssize_t d_stop = <int> f.shape[2] - 1 # note we do 1 less to stop AFTER we hit it
    cdef ssize_t d
    cdef np.ndarray[np.float_t, ndim=3] frame = f[:, :, ::1]
    
    # patch stuff
    cdef np.intp_t dx1 = dx + 1
    cdef np.intp_t dy1 = dy + 1
    cdef np.double_t patch_size = <float> (dx+dx1) * (dy+dy1)
    cdef np.float_t Rsq = R*R
    
    # size of samples/patches
    cdef np.intp_t n_samples = GOPT_FP.shape[0]
    cdef np.intp_t n_patches = GOPT_FP.shape[2]

    cdef np.ndarray[np.double_t, ndim=2] L = np.zeros((n_samples, n_patches), dtype=np.double)

    cdef np.intp_t i, j, ns, rx0, ry0, rx1, ry1, bsi, s, x, y
    cdef np.float_t dist, val, bdist
    cdef np.double_t BC

    cdef np.ndarray[np.float_t, ndim=2] pmodel
    cdef np.ndarray[np.float_t, ndim=1] pcounts
    
    # preallocate histogram to be length of largest number of samples
    j = 0
    for x in range(len(CC)):
        j += CC[x].shape[0]
    cdef np.ndarray[np.int_t, ndim=1] hst = np.zeros(j, dtype=np.int)
    
    # for each patch
    for j in range(n_patches):
        pmodel  = SS[j] # patch model samples - size ()
        pcounts = CC[j].astype('float') # patch sample counts
        
        ns = pcounts.shape[0]
        
        # check each candidate location for it - doing it this way avoids
        # accessing patch model and samples as often (only n_patches times)
        # and means we don't have to remake the temporary histogram array (hst)
        for i in range(n_samples):
            # patch location
            x = GOPT_FP[i, 0, j]
            y = GOPT_FP[i, 1, j]

            # patch bounds
            rx0 = x-dx ; ry0 = y-dy
            rx1 = x+dx1; ry1 = y+dy1

            # for each pixel
            for px in range(rx0, rx1):
                for py in range(ry0, ry1):
                    # best (matching) sample index reset to -1 (no match)
                    bsi = -1
                    bdist = Rsq

                    # for each sample
                    for s in range(ns):
                        
                        # calculate squared distance
                        dist = 0.0
                        d = 0
                        
                        while True:
                            val = pmodel[s, d] - frame[py, px, d]
                            dist += val * val
                            
                            # to avoid as many increments!
                            if d == d_stop:
                                break
                            
                            d += 1

                        # if it is < R**2 and the less than the current best distance..
                        if dist < Rsq:
                            if dist < bdist:
                                # it becomes the sample to match
                                bsi = s
                                bdist = dist

                    # if we got a match
                    if bsi > -1:
                        hst[bsi] += 1

            BC = 0.0
            for s in range(ns):
                BC += sqrt(hst[s] * pcounts[s])
                hst[s] = 0 
            
            # normalise by patch size so BC in [0, 1]
            # calculate Bhattacharyya distance (with minowski change)
            L[i, j] = 1.0 - pow(1.0 - BC / patch_size, power)
    
    return L
    
    
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)    
def locally_optimise_cand_patches_minkowski_floatcount(np.ndarray[np.intp_t, ndim=1] inds,
                               f, 
                               np.ndarray SS, 
                               np.ndarray CC, 
                               np.ndarray[np.int_t, ndim=3] GOPT_FP,
                               np.ndarray[np.float_t, ndim=2] GOPT_FP_likelihood,
                               np.intp_t dx, np.intp_t dy, np.int_t R,
                               np.intp_t n_LWS,
                               np.ndarray[np.uint8_t, ndim=2, cast=True] local_opt_mask=None,
                               np.double_t power=1.0):
    
    # number of feature channels (d is the channel no)
    cdef ssize_t d_stop = <int> f.shape[2] - 1 # note we do 1 less to stop AFTER we hit it
    cdef ssize_t d
    cdef np.ndarray[np.float_t, ndim=3] frame = f[:, :, ::1]
    
    # patch stuff
    cdef np.intp_t dx1 = dx + 1
    cdef np.intp_t dy1 = dy + 1
    cdef np.float_t patch_size = <float> (dx+dx1) * (dy+dy1)
    cdef np.float_t Rsq = R*R
    
    # size of samples/patches
    cdef np.intp_t n_samples = inds.shape[0]
    cdef np.intp_t n_patches = GOPT_FP.shape[2]
    
    # local optimisation window size
    cdef np.intp_t nx0 = -(n_LWS // 2)
    cdef np.intp_t ny0 = -(n_LWS // 2)
    
    cdef np.intp_t nx1 = (n_LWS // 2) + 1
    cdef np.intp_t ny1 = (n_LWS // 2) + 1
    
    # pointers of stuff
    cdef np.intp_t j, ns, i, x, y, rx0, ry0, rx1, ry1, sample_index, s, bsi
    cdef np.intp_t rx0o, rx1o, ry0o, ry1o, xo, yo, best_x, best_y, py, px, d1, d2
    
    cdef np.float_t dist, val, bdist, BC, BD, best_location_likelihood

    cdef np.ndarray[np.float_t, ndim=2] pmodel
    cdef np.ndarray[np.float_t, ndim=1] pcounts
    
    # preallocate histogram to be length of largest number of samples
    j = 0
    for x in range(len(CC)):
        j += CC[x].shape[0]
    cdef np.ndarray[np.int_t, ndim=1] hst = np.zeros(j, dtype=np.int)
    
    # check if local_opt_mask is None -> locally optimise all patches
    if local_opt_mask is None:
        local_opt_mask = np.ones((n_samples, n_patches), dtype=np.bool)
    
    # for each patch
    for j in range(n_patches):
        pmodel  = SS[j] # patch model samples - size ()
        pcounts = CC[j].astype('float') # patch sample counts
        
        ns = pcounts.shape[0]
        
        # for each candidate patch to be locally optimised
        for i in range(n_samples):
            # skip if we aren't locally optimising
            if not local_opt_mask[i, j]:
                continue
                
            sample_index = inds[i]
            
            # patch location
            x = GOPT_FP[sample_index, 0, j]
            y = GOPT_FP[sample_index, 1, j]

            # base patch bounds
            rx0 = x-dx ; ry0 = y-dy
            rx1 = x+dx1; ry1 = y+dy1
            
            # set negative likelihood to beat as best pos
            best_location_likelihood = 0
            best_x = 0
            best_y = 0
            
            # evaluate each possible move
            for xo in range(nx0, nx1):
                for yo in range(ny0, nx1):
                    # candidate local opt position
                    rx0o = rx0 + xo
                    rx1o = rx1 + xo
                    ry0o = ry0 + yo
                    ry1o = ry1 + yo
                    
                    # for each pixel
                    for py in range(ry0o, ry1o):
                        for px in range(rx0o, rx1o):
                            # best (matching) sample index reset to -1 (no match)
                            bsi = -1
                            bdist = Rsq

                            # for each sample
                            for s in range(ns):
                                # calculate squared distance
                                dist = 0.0
                                d = 0
                                while True:
                                    val = pmodel[s, d] - frame[py, px, d]
                                    dist += val * val
                                    
                                    if d == d_stop:
                                        break
                                        
                                    d += 1

                                # if it is < R**2 and the less than the current best distance..
                                if dist < Rsq:
                                    if dist < bdist:
                                        # it becomes the sample to match
                                        bsi = s
                                        bdist = dist

                            # if we got a match
                            if bsi > -1:
                                hst[bsi] += 1

                    # calc Bhattacharyya coef candidate patch vs its model and
                    # empty histogram for next candidate patch location
                    BC = 0.0
                    for s in range(ns):
                        BC += sqrt(hst[s] * pcounts[s])
                        hst[s] = 0 
                    BD = 1.0 - pow(1.0 - BC / patch_size, power)

                    # higher likelihood than the current best
                    if BD > best_location_likelihood:
                        best_x = xo
                        best_y = yo
                        best_location_likelihood = BD
                        
                    # equal likelihood
                    elif BD == best_location_likelihood:
                        # so decide based on distance from (0, 0)
                        d1 = xo*xo + yo*yo
                        d2 = best_x*best_x + best_y*best_y
                        if d1 < d2:
                            best_x = xo
                            best_y = yo
                            best_location_likelihood = BD
              
            # move patch to optimal location and save its likelihood
            GOPT_FP[sample_index, 0, j] += <int> best_x
            GOPT_FP[sample_index, 1, j] += <int> best_y
            GOPT_FP_likelihood[sample_index, j] = best_location_likelihood
