import numpy as np
from imageio import imread
from tracker_funcs import *

from patch_eval_funcs import (calc_cand_patch_likelihood_minkowski_floatcount,
                             locally_optimise_cand_patches_minkowski_floatcount)

class SBBM:
    def __param_init__(self):
        # show any numpy errors and ensure we have a new random seed for each
        # tracker instance (not guaranteed on Linux using threading otherwise)
        np.seterr(all='raise')
        np.random.seed(None)
        
        # patches
        self.patch_size = 5
        self.init_n_patches = 35
        self.n_samples = 10
        
        # matching
        self.R = 20
        self.minkowski_power = 1.4

        # global search, local search, local opt window size
        self.n_GAS = 1000
        self.n_LO = 100
        self.n_LWS = 5

        # model initialisation
        self.overlap_threshold = 0.25
        self.rho_plus = 1.5
        self.init_method = 'C'
        
        # updating model
        self.updaterate_counts = 0.05
        self.updaterate_samples = 1.7

        # new sampling scheme
        self.sd_x = 0.15
        self.sd_y = 0.1
        self.sd_r = np.pi / 16
        self.sd_s = 0.02
        
        self.mu_x = self.mu_y = self.mu_r = 0
        self.mu_s = 1        
        
    def __init__(self, frame, bbox):
        self.__param_init__()

        if type(frame) is str:
            frame = imread(frame)

        self.dx = self.dy = self.patch_size // 2
        self.ih, self.iw = frame.shape[:2]
        self.w, self.h = bbox_to_axis_aligned_bbox(bbox)[2:]
        self.bbox = bbox
        
        self.FP = patch_placement_superpixel(
            image=frame, bbox=bbox, dx=self.dx, dy=self.dy, 
            n_fp=self.init_n_patches, rho_plus=self.rho_plus, 
            overlap_threshold=self.overlap_threshold, method=self.init_method,
            debug=False, debug_final=False)

        self.samples = np.empty(self.FP.shape[1], dtype=object)
        self.counts = np.empty(self.FP.shape[1], dtype=object)

        # create model for each patch
        for i, (FP_x, FP_y) in enumerate(self.FP.T):
            region = [FP_x - self.dx, FP_y - self.dy,
                      FP_x + self.dx + 1, FP_y + self.dy + 1] # +1 as slicing
            s, c = create_model(frame, region, self.R, self.n_samples)
            
            self.samples[i] = s
            self.counts[i] = c

        # Gaussian info
        self.mean = np.array([0, 0, 0, 1], dtype='float')
        self.set_cov()
        
    def set_cov(self):
        t = 0.3 * (self.w + self.h) / 2
        self.cov = np.diag([t, t, np.pi/16, 0.02])**2
        
    def process_frame(self, frame):
        if type(frame) is str:
            frame = imread(frame)
            
        frame_float = frame.astype('float')

        # new sampling scheme
        sdx = self.sd_x * self.w
        sdy = self.sd_y * self.h
                                    
        GOPT_FP, GOPT_samples = generate_and_apply_affine_transforms_newscheme(
            self.FP, self.n_GAS, self.ih, self.iw, self.dx, self.dy, self.n_LWS, 
            sdx, sdy, self.sd_r, self.sd_s, self.mu_x, self.mu_y,
            self.mu_r, self.mu_s)

        # calculate patch likelihood
        GOPT_FP_likelihood = calc_cand_patch_likelihood_minkowski_floatcount(
                                    frame_float, self.samples, self.counts, 
                                    GOPT_FP, self.dx, self.dy, self.R, 
                                    power=self.minkowski_power)
        
        # model likelihood 
        GOPT_likelihood = np.mean(GOPT_FP_likelihood, axis=1)

        # select best to locally optimise
        opt_inds = np.argsort(GOPT_likelihood)[::-1][:self.n_LO]
        
        # perform local optimisation
        locally_optimise_cand_patches_minkowski_floatcount(opt_inds, 
                            frame_float, self.samples, self.counts, GOPT_FP, 
                            GOPT_FP_likelihood, self.dx, self.dy, self.R, 
                            self.n_LWS, power=self.minkowski_power)
                                                        
        # recalculate model likelihood for locally optimised patches
        GOPT_likelihood[opt_inds] = np.mean(GOPT_FP_likelihood[opt_inds], 
                                            axis=1)
        
        # highest likelihood model -> predicted location of object
        ind = np.argmax(GOPT_likelihood)
        self.FP = GOPT_FP[ind]
        
        # update patch models
        update_model_combined(frame_float, self.FP, self.samples, self.counts, 
                              self.dx, self.dy, 
                              alpha=self.updaterate_counts,
                              beta=self.updaterate_samples, 
                              R=self.R)
        
        # update additional info used in tracking       
        self.bbox = generate_bbox_from_FP(self.FP, self.dx, self.dy, method=0)
        self.w, self.h = bbox_to_axis_aligned_bbox(self.bbox)[2:]
        self.set_cov()
                     
        return self.bbox, self.FP, GOPT_FP_likelihood[ind]
 