import numpy as np

from alpha_matting_segmentation import perform_alpha_matting
from scipy.spatial import distance, ConvexHull
from skimage import measure, morphology, color, segmentation, draw

from pyDOE2.doe_lhs import lhs
from scipy.stats import laplace
from scipy.stats import norm as normal

def FP_outlier_replace_newver(outlier_mask, image, FP, dx, dy, 
                              rho_plus, overlap_threshold):
    # number of outliers
    n_outlier = np.sum(outlier_mask)

    # get bbox of points that aren't outliers
    non_outlier_bbox = generate_bbox_from_FP(FP[:, ~outlier_mask], 
                                             dx, dy, method=0)

    aa_bb = bbox_to_axis_aligned_bbox(non_outlier_bbox)

    # protect against failure due to tiny bbox by expanding it by
    # 1 pixel in each direction per failure
    failed = True

    while failed:
        try:
            cx0, cy0, cx1, cy1 = get_search_region(aa_bb, image, 
                                                   ratio=rho_plus)

            cimage = image[cy0:cy1, cx0:cx1].copy()
            cbbox = non_outlier_bbox.copy()
            cbbox[::2] -= cx0
            cbbox[1::2] -= cy0

            mask = alpha_matting_wrapper(cimage, cbbox)

            n_sp_approx = np.round(n_outlier / np.mean(mask)).astype('int')

            failed = False

        except:
            # move x0, y0 back and up by 1, increase h and w by 2
            aa_bb += [-1, -1, 2, 2]

    # superpixel the image
    segments = segmentation.slic(cimage, n_segments=n_sp_approx, slic_zero=True, 
                                 enforce_connectivity=True)

    segments_mask = segments.copy()
    segments_mask.flat[~mask.ravel()] = 0

    # calculate the properties (i.e. area and centroid) of each sp
    # note: labels with value 0 are ignored (hence setting background to 0)
    rps = measure.regionprops(segments_mask)

    # sort superpixels by area
    areas = np.array([_.area for _ in rps])
    inds = np.argsort(areas, )[::-1]

    cxx = [np.rint(rps[i].centroid[1]).astype('int') for i in inds]
    cyy = [np.rint(rps[i].centroid[0]).astype('int') for i in inds]

    # masks for each possible patch and the saved patches
    temp_mask = np.zeros_like(segments, dtype='bool')
    patch_mask = np.zeros_like(segments, dtype='bool')

    # number of pixels overlapping to count as too many
    n_pixel_thresh = (2*dx+1) * (2*dy+1) * overlap_threshold

    # temp variables
    overlap_count = np.zeros_like(cxx, dtype='int')
    temp_mask = np.zeros_like(segments, dtype='bool')
    patch_mask = np.zeros_like(segments, dtype='bool')

    # storage for new patches and counter for placed patches
    new_locations = np.zeros((2, n_outlier), dtype='int')
    nlc = 0 # new location counter

    # fill the patch mask with current patch locations
    for cx, cy in FP[:, ~outlier_mask].T:
        x0, y0, x1, y1 = cx-dx-cx0, cy-dy-cy0, cx+dx+1-cx0, cy+dy+1-cy0
        patch_mask[y0:y1, x0:x1] = True

    # repeat for each new patch needed
    while nlc < n_outlier:
        # fill in the patch mask for the new location(s) 
        # (if there are any so far)
        for i in range(nlc):
            cx, cy = new_locations[:, i]
            x0, y0, x1, y1 = cx-dx, cy-dy, cx+dx+1, cy+dy+1
            patch_mask[y0:y1, x0:x1] = True

        # find number of pixels overlapped for each location
        for i, (cx, cy) in enumerate(zip(cxx, cyy)):
            # raw pixel coords (adding one as only used for slicing here)
            x0, y0, x1, y1 = cx-dx, cy-dy, cx+dx+1, cy+dy+1

            # see if there is too much overlap
            temp_mask[:] = False
            temp_mask[y0:y1, x0:x1] = True
            overlap_count[i] = np.count_nonzero(temp_mask & patch_mask)

        # find those that overlap less than the threshold 
        mask_threshold_overlap = overlap_count <= n_pixel_thresh

        # if there are any below the threshold place a patch on the lowest
        # indexed one - as the corresponding sp will have highest area
        if np.any(mask_threshold_overlap):
            ind = np.where(mask_threshold_overlap)[0][0]

        # else get the one with the least overlap
        else:
            ind = np.argmin(overlap_count)

        # save location
        new_locations[:, nlc] = cxx[ind], cyy[ind]
        nlc += 1

    # translate new points to original frame dimensions
    new_locations += np.array([cx0, cy0])[:, np.newaxis]

    return new_locations

def place_FP_uniformly_across_bb(bbox, N, dx, dy):
    # get points and mean centre them
    x = bbox[::2].copy()
    y = bbox[1::2].copy()

    cx = np.mean(x)
    cy = np.mean(y)

    x -= cx
    y -= cy

    # calculate the angle of the rotated bb
    alpha = np.arctan2(x[0] - x[1], y[0] - y[1])

    # create rotation matrix to undo rotation
    R = np.array([[np.cos(alpha), -np.sin(alpha), 0],
                  [np.sin(alpha),  np.cos(alpha), 0],
                  [            0,              0, 1]])

    # matrix to REDO rotation
    Rinv = np.array([[np.cos(-alpha), -np.sin(-alpha), 0],
                     [np.sin(-alpha),  np.cos(-alpha), 0],
                     [             0,               0, 1]])

    # set up x and y in homogeneous coords
    xy1 = np.ones((3, x.size))
    xy1[0] = x
    xy1[1] = y

    # rotate back to unrotated bb
    rxy1 = R @ xy1

    # find its new bounds
    minx, maxx = np.min(rxy1[0]), np.max(rxy1[0])
    miny, maxy = np.min(rxy1[1]), np.max(rxy1[1])

    # round up the number of points to sample in each direction
    sqrtN = np.ceil(np.sqrt(N)).astype('int')

    # sample them
    xpoints = np.linspace(minx + dx + 1, maxx - dx - 1, sqrtN)
    ypoints = np.linspace(miny + dy + 1, maxy - dy - 1, sqrtN)
    xp, yp = np.meshgrid(xpoints, ypoints)

    # rotate them back
    xy2 = np.ones((3, xp.size))
    xy2[0] = xp.flat
    xy2[1] = yp.flat

    rxy2 = Rinv @ xy2

    # translate back to original coords
    FP = rxy2[:2, :] + np.array([[cx],[cy]])
    
    # remove random FPs so that we have N
    FP = FP[:, np.random.permutation(FP.shape[1])[:N]]

    return np.round(FP).astype('int')


def generate_and_apply_affine_transforms_newscheme(FP, n_GAS, ih, iw, dx, dy, 
                n_LWS, sd_x, sd_y, sd_r, sd_s, mu_x=0, mu_y=0, mu_r=0, mu_s=1):
    
    n_patches = FP.shape[1]
    
    GOPT_FP_store = np.zeros((n_GAS, 2, n_patches), dtype='int')
    GOPT_A = np.eye(3, dtype='float') # so GOPT_A[2, 2] = 1 always
    GOPT_FP_prev = np.ones((3, n_patches)) # so GOPT_FP_prev[2, :] = 1 always
    
    # sample from Gaussian
    GOPT_samples = generate_new_samples(n_GAS, sd_x, sd_y, sd_r, sd_s, 
                                        mu_x=mu_x, mu_y=mu_y, mu_r=mu_r, 
                                        mu_s=mu_s)

    # apply them to the previous timestep's FPs
    GOPT_FP_prev[:2] = FP
    GOPT_FP_prev_mean = np.mean(GOPT_FP_prev[:2], axis=1)

    for i in range(n_GAS):
        # create affine matrix as defined by the samples and FP's mean
        affine_transform(GOPT_A, GOPT_samples[i], GOPT_FP_prev_mean)

        # apply transform to FPs - [:2] as third row only needed to apply transform
        GOPT_FP_store[i] = np.rint((GOPT_A @ GOPT_FP_prev)[:2]).astype('int')

    # check which ones are valid (i.e. result in patches that lie inside the image)
    offset_x = dx+n_LWS+1 # patch width/height + local optimisation size + 1
    offset_y = dy+n_LWS+1
    valid_mask = np.any(GOPT_FP_store[:, 1, :] < offset_y, axis=1) # bottom, left
    valid_mask |= np.any(GOPT_FP_store[:, 0, :] < offset_x, axis=1) # bottom, left
    valid_mask |= np.any(GOPT_FP_store[:, 1, :] > ih-1-offset_y, axis=1) # top
    valid_mask |= np.any(GOPT_FP_store[:, 0, :] > iw-1-offset_x, axis=1) # right

    invalid_inds = np.where(valid_mask)[0]
    
    stg = np.maximum(len(invalid_inds), 100) # number of Samples To Generate
    c = stg # random sample pointer 
              # (starts at stg so we generate samples IN the loop)

    # replace the invalid sets of FP locations
    for i in invalid_inds:
        while True:
            # if we've ran out of samples then generate some more!
            if c == stg:
                random_samples = generate_new_samples(stg, sd_x, sd_y, sd_r,
                                      sd_s, mu_x=mu_x, mu_y=mu_y, mu_r=mu_r,
                                      mu_s=mu_s)
                c = 0
            
            # get a random sample, increase pointer
            GOPT_samples[i] = random_samples[c]
            c += 1
            
            # apply the transform and store it
            affine_transform(GOPT_A, GOPT_samples[i], GOPT_FP_prev_mean)
            GOPT_FP_store[i] = np.rint((GOPT_A @ GOPT_FP_prev)[:2]).astype('int')

            # test if its valid, repeat if not
            if not (np.any(GOPT_FP_store[i, 1, :] < offset_x)
                    or np.any(GOPT_FP_store[i, 0, :] < offset_y)
                    or np.any(GOPT_FP_store[i, 1, :] > ih-1-offset_x)
                    or np.any(GOPT_FP_store[i, 0, :] > iw-1-offset_y)):
                break
                
    return GOPT_FP_store, GOPT_samples


def generate_new_samples(n_samples, sd_x, sd_y, sd_r, sd_s, 
                         mu_x=0, mu_y=0, mu_r=0, mu_s=1):
    # 1. generate LHS
    S = lhs(4, samples=n_samples, criterion='m') # shape = (n_samples, 4)

    # 2. Transform into Laplacian or Normal distributions with specified
    #    statistics
    S[:, 0] = reverse_Laplacian(S[:, 0], sd=sd_x, mu=mu_x)
    S[:, 1] = reverse_Laplacian(S[:, 1], sd=sd_y, mu=mu_y)
    S[:, 2] = reverse_Normal   (S[:, 2], sd=sd_r, mu=mu_r)
    S[:, 3] = reverse_Normal   (S[:, 3], sd=sd_s, mu=mu_s)
    
    return S

    
def reverse_Laplacian(y, sd, mu=0.0):
    return laplace.ppf(y, loc=mu, scale=sd)

    
def reverse_Normal(y, sd, mu=0.0):
    return normal.ppf(y, loc=mu, scale=sd)
    
    
def update_model_combined(image, FP, S, C, dx, dy,
                          alpha, 
                          beta,
                          R, outlier_mask=None,
                          normalise_counts=False,
                          closefar=None):

    nd = image.shape[2]
    Rsq = R * R
    patch_area = (2 * dx + 1) * (2 * dy + 1)

    for i, (FP_x, FP_y) in enumerate(FP.T):
        # if we have an outlier mask and this patch is an outlier, skip it
        if (outlier_mask is not None) and outlier_mask[i]:
            continue

        # patch inds (for slicing, hence +1)
        x0, x1, y0, y1 = FP_x-dx, FP_x+dx+1, FP_y-dy, FP_y+dy+1

        # select the patch we're working on
        patch = np.reshape(image[y0:y1, x0:x1, :], (-1, nd))

        # sqrd euc dist between patch and ALL samples (patchsize**2, n_samples)
        dists = np.sum((patch[:, np.newaxis] - S[i])**2, axis=2)

        # mask of matches
        matches = np.any(dists < Rsq, axis=1)

        # separate the patch into two, one with pixels that match samples
        # and one without
        patch_match = patch[matches]
        patch_nomatch = patch[~matches]

        # get the number of pixels that match each current sample
        # and update the sample's location to be a convex combination
        # of itself and the mean of the matches

        # indices of closest sample that lies within R
        am = np.argmin(dists[matches, :], axis=1)

        n_samples = S[i].shape[0]
        
        C_match = np.zeros((n_samples, ), dtype='float')
        S_match_mean = np.zeros((n_samples, nd), dtype='float')

        for j in range(n_samples):
            m = patch_match[am == j, :]

            # if we have any matches for a sample
            if m.shape[0] != 0:
                # smooth the sample
                S[i][j, :] *= 1 - beta
                
                # store the mean of the matching pixels
                S_match_mean[j] = np.mean(m, axis=0)
                
                # count how many matches
                C_match[j] = m.shape[0]

        # update match counts of current samples and their locations
        C[i] = (1-alpha) * C[i] + alpha * C_match
        S[i] += (beta * S_match_mean)

        # for all remaining pixels, i.e. those that did not match a sample,
        # employ the model creation scheme of choice
        n_nomatch = patch_nomatch.shape[0]

        if n_nomatch > 0:
            # nomatches will have (n_nomatch, 3) shape, so set region to eval
            # as x = [0, n_nomatch], y = 0
            region = [0, 0, n_nomatch, 1]
            
            # add fake axis so we can index y (just 0), to fit in with existing
            # model creation functions and ensure we don't cut off any samples
            # by having max allowed as n_nomatch (i.e. total amount)
            
            if closefar == None:
                new_s, new_c = create_model(patch_nomatch[np.newaxis, :, :], region, 
                                        R=R, n_samples=n_nomatch)
            elif closefar == 'close':
                new_s, new_c = create_model_nearfar(
                                        patch_nomatch[np.newaxis, :, :], region, 
                                        n_samples=n_nomatch, R=R, closest=True)
            elif closefar == 'far':
                new_s, new_c = create_model_nearfar(
                                        patch_nomatch[np.newaxis, :, :], region, 
                                        n_samples=n_nomatch, R=R, closest=False)
                                        
            else:
                raise Exception(closefar)

            # combine the new samples with the old
            S[i] = np.concatenate((S[i], new_s))

            # add in the new sample's counts, weighted with the update rate
            C[i] = np.concatenate((C[i], alpha * new_c))
         
        # remove samples/counts with a count of less than alpha
        inds = np.where(C[i] < alpha)[0]

        if inds.shape[0] > 0:
            C[i] = np.delete(C[i], inds)
            S[i] = np.delete(S[i], inds, axis=0)
        
        # re-normalise so that counts sum to area of patch
        if normalise_counts:
            C[i] = (C[i] / np.sum(C[i])) * patch_area


def scale_colour_channels_to_match_radii(image, R_L, R_AB):
    # we want to be able to use the same radius for all 3 channels (R_AB)
    # but have L be scaled such that by using the radius R_AB in the L
    # channel we are effectively using R_L
    image[:, : ,0] *= ( R_AB / R_L )
    return image

    
def my_rgb2hsv(I, ignore_intensity=False):
    I = color.rgb2hsv(I)
    
    # scale to [0, 255]
    I *= 255
    
    if ignore_intensity:
        return I[:, :, :2]
    
    return I


def my_rgb2lab(I, ignore_intensity=False):
    I = color.rgb2lab(I)
    
    # skipping rescaling to try leaving as-is
    if False:
        ### these normalisation constants taken from largest/smallest
        #   possible LAB values for any integer rgb value in [0, 0, 0]
        #   to [255, 255, 255]
        
        # normalise L from [0, 100] to [0, 1]
        I[:, :, 0] /= 100 
        
        # normalise A from [-86.18302974, 98.23305386] to [0, 1]
        I[:, :, 1] += 86.18302974
        I[:, :, 1] /= (86.18302974 + 98.23305386)
        
        # normalise B from [-107.85730021, 94.47812228] to [0, 1]
        I[:, :, 2] += 107.85730021
        I[:, :, 2] /= (107.85730021 + 94.47812228)
        
        # scale to [0, 255]
        I *= 255
    
    if ignore_intensity:
        return I[:, :, 1:]
    
    return I

    
def update_tracking_model_newver(image, samples, FP, dx, dy, 
                                 alpha=0.5, R=20, outlier_mask=None):
    n_patches = samples.shape[0]
    nd = image.shape[2]

    Rsq = R**2

    for i, (FP_x, FP_y) in enumerate(FP.T):
        # if we have an outlier mask and this patch is an outlier, skip it
        if (outlier_mask is not None) and outlier_mask[i]:
            continue

        # patch inds (for slicing, hence +1)
        x0, x1, y0, y1 = FP_x-dx, FP_x+dx+1, FP_y-dy, FP_y+dy+1

        # select the patch we're working on
        patch = np.reshape(image[y0:y1, x0:x1, :], (-1, nd))

        # sqrd euc dist between patch and ALL samples (patchsize**2, n_samples)
        dists = np.sum((patch[:, np.newaxis] - samples[i])**2, axis=2)

        # mask of matches
        matches = np.any(dists < Rsq, axis=1)

        # indices of closest sample that lies within R
        am = np.argmin(dists[matches, :], axis=1)

        # limit patch to only contain pixels that match a sample
        patch = patch[matches]

        # update each sample to be a weighted some of itself and its matches
        for j in range(samples[i].shape[0]):
            # pixel values matches that lie within R of sample j
            m = patch[am == j, :]

            # if we have any matches, smoothly update the sample centre
            if m.shape[0] != 0:
                mn = np.mean(m, axis=0)
                samples[i][j, :] =  (1-alpha)*samples[i][j, :] + alpha * mn


def create_model_nearfar(image, region, n_samples=10, R=20, 
                         closest=False):
    x0, y0, x1, y1 = region
    Rsq = R**2

    # flatten patch so that it is (n_pixels by feature dimension)
    # we need the float else we get errors when subtracting!
    patch = np.reshape(image[y0:y1, x0:x1, :], 
                            (-1, image.shape[2])).astype('float')
    n_px, d = patch.shape

    ######### STORAGE
    # location of each sample
    samples = np.zeros((n_px, d), dtype='float')
    # mask of pixels we still need to assign
    remaining_mask = np.ones((n_px, ), dtype='bool')
    # sample counter
    c = 0

    # if we're dealing with the closest, pick the pixel nearest the centre
    if closest:
        centroid = np.mean(patch, axis=0)
        ind = np.argmin(np.linalg.norm(patch - centroid, axis=1))
    # else we pick the pixel nearest the origin -> i.e. on the edge
    else:
        ind = np.argmin(np.linalg.norm(patch, axis=1))

    # store first sample
    samples[c] = patch[ind]    
    remaining_mask[ind] = False
    c += 1

    while True:
        # calculate distance (squared euc) from samples to all remaining pixels
        temp = np.sum((samples[:c, np.newaxis, :] - 
                              patch[np.newaxis, remaining_mask, :])**2, axis=2)

        # find each pixel's distance the sample closest
        D = np.min(temp, axis=0)
        
        # mask out pixels that cannot become a new sample point
        # because they already match an existing sample
        match_mask = D < Rsq
        D = D[~match_mask]
        
        # remove matching pixels those considered as a sample location
        remaining_mask[np.where(remaining_mask)[0][match_mask]] = False
        
        # check to see if any pixels left that do not match a sample
        if D.shape[0] == 0:
            break

        # find the pixel that is closest/farthest from its nearest sample
        ind = np.argmin(D) if closest else np.argmax(D)

        # store the sample and increase the sample counter
        samples[c] = patch[remaining_mask][ind]
        c += 1
        
    samples = samples[:c, :]
    
    # number of pixels in the patch that match it
    counts = np.zeros((c, ), dtype='int')
    
    # calculate distance from each sample to the original patch
    D = np.sum((samples[:, np.newaxis, :] - patch[np.newaxis, :, :])**2, axis=2)
    matches = D < Rsq
    n_matches_per_px = np.count_nonzero(matches, axis=0)
    
    for i in range(n_px):
        # if a pixel matches exactly one sample
        if n_matches_per_px[i] == 1:
            counts[matches[:, i]] += 1
            
        # else there must be multiple matches (there can never be 
        # zero matches a sample will match at least itself)
        else:
            counts[np.argmin(D[:, i])] += 1
    
    # sort the counts by amount of pixels (highest to lowest)
    # and keep only the 'n_samples' samples with the highest counts
    sorted_inds = np.argsort(counts)[::-1][:n_samples]
    
    return samples[sorted_inds], counts[sorted_inds]
                
                
def create_model_nearfar_old(image, region, n_samples=10, R=20, 
                         closest=False, return_matches=False):
    x0, y0, x1, y1 = region
    Rsq = R**2

    # flatten patch so that it is (n_pixels by feature dimension)
    # we need the float else we get errors when subtracting!
    orig_patch = np.reshape(image[y0:y1, x0:x1, :], 
                            (-1, image.shape[2])).astype('float')
    n_px, d = orig_patch.shape

    ######### STORAGE
    # location of each sample
    samples = np.zeros((n_px, d), dtype='float')
    # number of pixels in the patch that match it
    counts = np.zeros((n_px, ), dtype='int')
    # mask of pixels that match it
    matches = np.zeros((n_px, n_px), dtype='bool')
    # mask of pixels we still need to assign
    remaining_mask = np.ones((n_px, ), dtype='bool')
    # sample counter
    c = 0

    # if we're dealing with the closest, pick the pixel nearest the centre
    if closest:
        centroid = np.mean(orig_patch, axis=0)
        ind = np.argmin(np.linalg.norm(orig_patch - centroid, axis=1))
    # else we pick the pixel nearest the origin -> i.e. on the edge
    else:
        ind = np.argmin(np.linalg.norm(orig_patch, axis=1))

    # update samples, matches, and counts
    samples[c] = orig_patch[ind]
    matches[c] = np.sum((orig_patch - orig_patch[ind])**2, axis=1) < Rsq
    counts[c] = np.count_nonzero(matches[c])

    # update remaining pixels left to match and counter
    remaining_mask &= ~matches[c]
    patch = orig_patch[remaining_mask]
    c += 1

    # while there ate still pixels to assign..
    while patch.shape[0] > 0:
        # calculate from centroids to all remaining pixels
        temp = np.sum((samples[:c, np.newaxis, :] - patch[np.newaxis, :, :])**2, axis=2)
        # find the least distance from a centroid
        D = np.min(temp, axis=0)

        # either find the closest or furthest away (outside of Rsq away) to centroids
        ind = np.argmin(D) if closest else np.argmax(D)

        # store the sample, its matches, and count
        samples[c] = patch[ind]
        matches[c][remaining_mask] = np.sum((patch - patch[ind])**2, axis=1) < Rsq
        counts[c] = np.count_nonzero(matches[c])

        # update remaining pixels left to match and counter
        remaining_mask &= ~matches[c]
        patch = orig_patch[remaining_mask]
        c += 1

    samples = samples[:c, :]
    counts = counts[:c]
    matches = matches[:c, :]
        
    # sort the counts by amount of pixels (highest to lowest)
    # and keep only the 'n_samples' samples with the highest counts
    sorted_inds = np.argsort(counts)[::-1][:n_samples]

    samples = samples[sorted_inds, :]
    counts = counts[sorted_inds]

    if not return_matches:
        return samples, counts

    matches = matches[sorted_inds, :]
    
    return samples, counts, matches


def create_model_kmeansplusplus(image, region, n_samples=10, 
                                n_iter=100, R=20, return_matches=False):
    x0, y0, x1, y1 = region
    Rsq = R**2
    
    # flatten patch so that it is (n_pixels by feature dimension)
    # we need the float else we get errors when subtracting!
    orig_patch = np.reshape(image[y0:y1, x0:x1, :], 
                            (-1, image.shape[2])).astype('float')
    n_px, d = orig_patch.shape

    ######### STORAGE
    # location of each sample
    samples = np.zeros((n_px, d), dtype='float')
    # number of pixels in the patch that match it
    counts = np.zeros((n_px, ), dtype='int')
    # mask of pixels that match it
    matches = np.zeros((n_px, n_px), dtype='bool')
    # mask of pixels we still need to assign
    remaining_mask = np.ones((n_px, ), dtype='bool')
    # number of samples
    c = 0
    ##########################################

    best_k = np.inf
    
    for _ in range(n_iter):
        # reset everything
        c = 0
        samples.fill(0)
        counts.fill(0)
        matches.fill(False)
        remaining_mask.fill(True)
        
        # select initial location randomly
        ind = np.random.randint(n_px)

        # store sample, matches, and count
        samples[c] = orig_patch[ind]
        matches[c] = np.sum((orig_patch - orig_patch[ind])**2, axis=1) < Rsq
        counts[c] = np.count_nonzero(matches[c])

        # update remaining pixels left to match and counter
        remaining_mask &= ~matches[c]
        patch = orig_patch[remaining_mask]
        c += 1

        # while there are still pixels to assign
        while patch.shape[0] > 0:
            #### k-means++ selection of next sample ####
            # find distance of all remaining pixels to all samples
            temp = samples[:c, np.newaxis, :] - patch[np.newaxis, :, :]
            D = np.min(np.sum(temp**2, axis=2), axis=0)

            # choose next ind such that likelihood of choosing it is 
            # proportional to the closest existing cluster centre
            probs = D / np.sum(D)
            cprobs = np.cumsum(probs)
            ind = np.where(cprobs >= np.random.random())[0][0]
            ############################################

            # store the sample, its matches, and count
            samples[c] = patch[ind]
            matches[c][remaining_mask] = np.sum((patch - patch[ind])**2, axis=1) < Rsq
            counts[c] = np.count_nonzero(matches[c])

            # update remaining pixels left to match and counter
            remaining_mask &= ~matches[c]
            patch = orig_patch[remaining_mask]
            c += 1
        
        if (c-1) < best_k:
            best_samples = samples[:c, :].copy()
            best_counts = counts[:c].copy()
            best_matches = matches[:c, :].copy()
            best_k = c-1
    
    # sort the counts by amount of pixels (highest to lowest)
    # and keep only the 'n_samples' samples with the highest counts
    sorted_inds = np.argsort(best_counts)[::-1][:n_samples]

    best_samples = best_samples[sorted_inds, :]
    best_counts = best_counts[sorted_inds]
    
    if not return_matches:
        return best_samples, best_counts
    
    best_matches = best_matches[sorted_inds, :]
    return best_samples, best_counts, best_matches


def update_tracking_model(image, samples, counts, FP, dx, dy, alpha=0.5, R=20,
                          outlier_mask=None):
    n_patches, n_samples, _ = samples.shape

    Rsq = R**2

    for i, (FP_x, FP_y) in enumerate(FP.T):
        # if we have an outlier mask and this patch is an outlier, skip it
        if (outlier_mask is not None) and outlier_mask[i]:
            continue

        # patch inds (for slicing, hence +1)
        x0, x1, y0, y1 = FP_x-dx, FP_x+dx+1, FP_y-dy, FP_y+dy+1

        # select the patch we're working on
        patch = np.reshape(image[y0:y1, x0:x1, :], (-1, 3))

        # sqrd euc dist between patch and ALL samples (patchsize**2, n_samples)
        dists = np.sum((patch[np.newaxis, :, :] - 
                        samples[i, :, np.newaxis])**2, axis=2)

        # mask of matches
        matches = np.any(dists < Rsq, axis=0)

        # indices of closest sample that lies within R
        am = np.argmin(dists, axis=0)[matches]

        # limit patch to only contain pixels that match a sample
        patch = patch[matches]

        for j in range(n_samples):
            # pixel values matches that lie within R of sample j
            m = patch[am == j, :]

            # if we match, smoothly update the sample centre
            if m.shape[0] != 0:
                mn = np.mean(m, axis=0)
                samples[i, j, :] =  (1-alpha)*samples[i, j, :] + alpha * mn

def minBoundingBox(X):
    assert(X.shape[0] == 2)
    
    k = ConvexHull(X.T)

    # convex hull of FP
    k = np.append(k.vertices, k.vertices[0])

    CH = X[:, k]

    # compute the angle to test, which are the angle of the CH edges as:
    # "one side of the bounding box contains an edge of the convex hull"
    E = np.diff(CH, n=1, axis=1) # ch edges
    T = np.arctan2(E[1], E[0]) # angles of edges
    T = np.unique(np.mod(T, np.pi/2)) # unique set of 1st quadrant angles

    # create rotation matrix which contains
    # the 2x2 rotation matrices for *all* angles in T
    # R is a 2n*2 matrix
    p1 = np.tile(np.tile(T, (2, 1)).T.ravel(), (2, 1)).T
    p2 = np.tile(np.array([[0, -np.pi], [np.pi, 0]])/2, (len(T), 1))
    R = np.cos(p1 + p2)

    # rotate CH by all angles
    RCH = R @ CH

    # compute border size  [w1;h1;w2;h2;....;wn;hn]
    # and area of bounding box for all possible edges
    bsize = np.max(RCH, axis=1) - np.min(RCH, axis=1)
    area = np.prod(np.reshape(bsize, (-1, 2)), axis=1)

    # ind of minimum area -> angle in T
    i = np.argmin(area)

    # compute the bound (min and max) on the rotated frame
    Rf = R[2*i:2*i+2, :] # rotated frame
    bound = Rf @ CH
    bmin = np.min(bound, axis=1)
    bmax = np.max(bound, axis=1)

    Rf = Rf.T;
    bb = np.zeros(8)
    bb[0:2] = bmin[0]*Rf[:, 0] + bmin[1]*Rf[:, 1]
    bb[2:4] = bmin[0]*Rf[:, 0] + bmax[1]*Rf[:, 1]
    bb[4:6] = bmax[0]*Rf[:, 0] + bmax[1]*Rf[:, 1]
    bb[6:8] = bmax[0]*Rf[:, 0] + bmin[1]*Rf[:, 1]
    
    return bb
    
    
def order_bbox_corners(bb):
    bb_ordered = np.zeros(8)

    I = np.argsort(bb[1::2])

    # top-left and top-right
    if bb[2*I[0]] < bb[2*I[1]]:
        bb_ordered[0:2] = bb[2*I[0]:2*I[0]+2]
        bb_ordered[2:4] = bb[2*I[1]:2*I[1]+2]
    else:
        bb_ordered[0:2] = bb[2*I[1]:2*I[1]+2]
        bb_ordered[2:4] = bb[2*I[0]:2*I[0]+2]

    if bb[2*I[2]] > bb[2*I[3]]:
        bb_ordered[4:6] = bb[2*I[2]:2*I[2]+2]
        bb_ordered[6:8] = bb[2*I[3]:2*I[3]+2]
    else:
        bb_ordered[4:6] = bb[2*I[3]:2*I[3]+2]
        bb_ordered[6:8] = bb[2*I[2]:2*I[2]+2]

    return bb_ordered
    
    
def expand_ordered_bbox(bb, dx, dy):
    # assuming bb[:2] = top left, bb[2:4] = top right,
    # bb[4:6] = bot right, bb[6:8] = bot left
#     bb_expanded = np.zeros(8)
    
#     bb_expanded[0:2] = bb[0:2] + [-dx, -dy]
#     bb_expanded[2:4] = bb[2:4] + [ dx, -dy]
#     bb_expanded[4:6] = bb[4:6] + [ dx,  dy]
#     bb_expanded[6:8] = bb[6:8] + [-dx,  dy]
    
    return bb + [-dx, -dy, dx, -dy, dx, dy, -dx, dy]


def get_neighbourhood_inds(w, h, cx=0, cy=0):
    # returns x and y coords of all neighbours in the (h, w) window
    # centred on (cx, cy)
    # NOTE: includes cx, cy.
    if w % 2 != 1 or h % 2 != 1:
        raise ValueError(f'w ({w:d}) and h ({h:d}) must be odd')
    w = w//2
    h = h//2
    return [_.ravel() for _ in np.mgrid[cx-w:cx+w+1, cy-h:cy+h+1]]


def get_search_region(bbox, frame, ratio):
    """
    Calculates coordinates of ratio*width/height of axis-aligned bbox, centred
    on the original bbox, constrained by the original size of the image.
    
    Arguments:
      bbox  = axis-aligned bbox of the form [x0, y0, width, height]
      frame = MxNxD Image to constrain bbox by
      ratio = ratio at which to change bbox dimensions by
      
    Output:
      x0, y0, x1, y1 = Coordinates of expanded axis-aligned bbox
    """
    x0, y0, w, h = bbox
    ih, iw = frame.shape[:2]
    
    ww, hh = ratio*w, ratio*h
    
    # expand bbox by ratio
    x1 = np.min([iw-1, x0 + w/2 + ww/2])
    y1 = np.min([ih-1, y0 + h/2 + hh/2])
    x0 = np.max([0, x0 + w/2 - ww/2])
    y0 = np.max([0, y0 + h/2 - hh/2])
    
    x0, y0, x1, y1 = np.rint(np.array([x0, y0, x1, y1])).astype('int')
    
    return x0, y0, x1, y1


def alpha_matting_wrapper(image, bbox, rho_plus=1.1, rho_minus=0.9, 
                          lbda=1e-2, tau=0.5):
    # perform alpha matting to segment the object in the bbox
    mask = perform_alpha_matting(image, bbox, rho_plus=rho_plus, 
                                 rho_minus=rho_minus, lbda=lbda, tau=tau)

    # label each disconnected region of the binary mask
    mask_label = measure.label(mask) 

    # calculate properties of each labelled region - ignores 0'th label
    # such that r[0] has the label 1, r[1] -> 2 etc..
    regions = measure.regionprops(mask_label) 

    # find the region with the largest area and remove all others
    areas = [r['area'] for r in regions]
    label_max_area = np.argmax(areas) + 1
    mask_label.flat[mask_label.flat != label_max_area] = 0

    # replace original mask with only largest region left
    # note: bool maps any label > 0 to 1
    return np.array(mask_label, dtype='bool')
    

def patch_placement_superpixel(image, bbox, dx, dy, n_fp, rho_plus,
                               overlap_threshold=0, method='C',
                               debug=False, debug_final=False,
                               use_entire_bb_as_object=False):

    if method not in ['C', 'E']:
        raise ValueError(f'Method ({method:s}) must be "E" or "C" ')
    
    aa_bb = bbox_to_axis_aligned_bbox(bbox)
    
    # protect against failure due to tiny bbox by expanding it by
    # 1 pixel in each direction per failure
    failed = True
    
    while failed:
        
        try:
            cx0, cy0, cx1, cy1 = get_search_region(aa_bb, image, ratio=rho_plus)

            cimage = image[cy0:cy1, cx0:cx1].copy()
            cbbox = bbox.copy()
            cbbox[::2] -= cx0
            cbbox[1::2] -= cy0

            # if we're not using alpha matting, create an mask and fill in
            # True where the bounding box lies
            if use_entire_bb_as_object:
                mask = np.zeros((cimage[:, :, 0].shape), dtype='bool')
                h, w, = cimage[:, :, 0].shape
                rows, cols = draw.polygon(cbbox[::2], cbbox[1::2], (w, h))
                mask[cols, rows] = True
                
            # else use alpha matting
            else:
                mask = alpha_matting_wrapper(cimage, cbbox)
            
            # calculate number of superpixels to approximately make. 
            # done by comparing the ratio of segmented image to total area 
            # of the image and assuming we want n_fp to lie within 
            # the segmented region
            n_sp_approx = np.round(n_fp / np.mean(mask)).astype('int')
            
            failed = False
        except:
            # move x0, y0 back and up by 1, increase h and w by 2
            aa_bb += [-1, -1, 2, 2]
        


    # superpixel the image
    segments = segmentation.slic(cimage, n_segments=n_sp_approx, slic_zero=True, 
                    enforce_connectivity=True)

    # set all pixels that lie outside the segmented region to have label 0
    segments_mask = segments.copy()
    segments_mask.flat[~mask.ravel()] = 0
    
    if method == 'C':
        # calculate the properties (i.e. area and centroid) of each sp
        # note: labels with value 0 are ignored (hence setting background to 0)
        rps = measure.regionprops(segments_mask)

        # sort superpixels by area
        areas = np.array([_.area for _ in rps])
        inds = np.argsort(areas, )[::-1]
        
        cxx = [np.rint(rps[i].centroid[1]).astype('int') for i in inds]
        cyy = [np.rint(rps[i].centroid[0]).astype('int') for i in inds]
    
    elif method == 'E':
        # mask of superpixel boundaries
        boundaries = segmentation.find_boundaries(segments)

        # get coordinates of pixels that lie on the boundary of sps and within mask
        y, x = np.where(boundaries & mask)

        # indices of pixels lying with a (w, h) window centred on (0, 0)
        xn, yn = get_neighbourhood_inds(2*dx+1, 2*dy+1)

        # combine these so we sample them centred on each of the pixels, i.e.
        # yy[i] = y[i] + yn, so yy.shape = (len(y), len(yn))
        yy = yn + y[:, np.newaxis]
        xx = xn + x[:, np.newaxis]
        yy = np.maximum(yy, 0); yy = np.minimum(yy, segments.shape[0]-1)
        xx = np.maximum(xx, 0); xx = np.minimum(xx, segments.shape[1]-1)

        # count the number of unique labels within each pixel's neighbourhood
        neighbour_counts = np.array([len(np.unique(row)) for row in segments[yy, xx]])

        inds = np.argsort(neighbour_counts)[::-1] # most to least unique
        cxx, cyy = x[inds], y[inds]
    
    # masks for each possible patch and the saved patches
    temp_mask = np.zeros_like(segments, dtype='bool')
    patch_mask = np.zeros_like(segments, dtype='bool')

    # number of pixels overlapping to count as too many
    n_pixel_thresh = (2*dx+1) * (2*dy+1) * overlap_threshold

    FP = np.zeros((2, n_fp), dtype='int')
    
    n_patches = 0
    for cx, cy in zip(cxx, cyy):
        # raw pixel co-ords (adding one as only used for slicing here)
        x0, y0, x1, y1 = cx-dx, cy-dy, cx+dx+1, cy+dy+1

        # see if there is too much overlap
        temp_mask[:] = False
        temp_mask[y0:y1, x0:x1] = True
        overlap_count = np.count_nonzero(temp_mask & patch_mask)
        
        # if there isn't then add it to the patch mask and FP
        if overlap_count <= n_pixel_thresh:
            patch_mask[y0:y1, x0:x1] = True
            FP[:, n_patches] = [cx, cy]
            n_patches += 1
            
            # stop if we've gotten enough FPs
            if n_patches == n_fp:
                break

    # slice off any unused patch locations 
    FP = FP[:, :n_patches]
    
    # move back from relative to actual coordinates
    FP += np.rint([cx0, cy0]).astype('int')[:, np.newaxis]
    
    # clip to image boundaries
    FP[0] = np.minimum(np.maximum(FP[0], dx), image.shape[1] - 1 - dx)
    FP[1] = np.minimum(np.maximum(FP[1], dy), image.shape[0] - 1 - dy)
    
    return FP

    
def wrap_angle(a, lower=0, upper=2 * np.pi):
    # wraps angle a to lie in [lower, upper)

    step = upper if a < lower else -upper
    while not (lower <= a < upper):
        a += step

    return a


def angle_fix(a, b, pi=np.pi):
    # if we want to update a like: a = (1-alpha)*a + alpha*b
    # change b such that a will always move towards the bearing
    # b faces. i.e. if a = 20 and b = 340, the sum above would
    # increase a's amount when we want it decreased, hence the
    # function would change b to -20 so that a moves in the
    # direction of b

    # ensure a, b in [0, 2pi)
    a = wrap_angle(a)
    b = wrap_angle(b)

    # the code below equates to the following:
    # if a < b and abs(a-b) <= pi then do nothing
    # if a < b and abs(a-b) > pi then a must be in right half and b
    #  the left half, so a wants to rotate left towards b, hence b becomes
    #  the negative bearing
    #
    # if a > b and abs(a-b) <= pi also do nothing
    # if a > b and abs(a-b) > pi then a must be in left half and b in
    #   the right half, so want to rotate a right towards b, but as b is
    #   smaller than a, we add 2pi to make it larger but same bearing to
    #   steer in correct direction..
    # i.e:
    # if a > b and abs(a - b) > pi:
    #    b = b + 2 * pi
    # elif a < b and abs(a - b) > pi:
    #    b = b - 2 * pi

    if np.abs(a - b) > pi:
        b += -2 * pi if a < b else 2 * pi

    return b

    
def calculate_hog_features(I, out=None, angles=None, magnitude=None):
    Igray = color.rgb2gray(I)
    
    # hog storage
    if out is None:
        out = np.zeros((*Igray.shape, 9), dtype='float64')
    # if we're reusing an array then we need to empty it as we only put magnitudes into certain bins
    else:
        out[:] = 0
        
    if angles is None:
        angles = np.zeros_like(Igray)
    if magnitude is None:
        magnitude = np.zeros_like(Igray)

    # calculate gradients, angles and magnitudes
    Ix = np.zeros_like(Igray)
    Iy = np.zeros_like(Igray)

    Ix[:, 1:-1] = (Igray[:, 2:] - Igray[:, :-2])/2
    Iy[1:-1, :] = (Igray[2:, :] - Igray[:-2, :])/2
    
    # arctan2 = -pi to pi -> signed bins. +1e-8 to avoid div by 0 errors
    angles[:] = np.arctan2(Iy, (Ix + 1e-8))
    magnitude[:] = (Ix**2 + Iy**2)**0.5

    # for each pixel, place its magnitude into the corresponding angle bin
    angle_steps = np.linspace(-np.pi, np.pi, 9+1)

    for i in range(9):
        lower, upper = angle_steps[i], angle_steps[i+1]
        if i == 0:
            mask = (lower <= angles) & (angles <= upper)
        else:
            mask = (lower < angles) & (angles <= upper)
        out[:, :, i].flat[mask.ravel()] = magnitude.flat[mask.ravel()]
        
    #assert np.allclose(out.sum(), magnitude.sum())
    return out

    
def update_model(samples, counts, frame, FP, dx, dy, update_rate=0.05, R=20, n_samples=10):
    # samples   = previous timestep's samples
    # counts    = previous timestep's counts
    # frame     = image to work on
    # FP        = 2 x number of FP x,y coords s.t. FP[0] = x_1, y_1
    # dx, dy    = half-lengths of the FPatches, so a patch's x-coord is:
    #             FP[0]-dx to FP[0]+dx(+1) <- +1 in brackets needed if slicing
    # update_rate = update rate of model, smoothing param
    # R         = vibe-like radius for a match
    # n_samples = total number of samples to use (clips to this if more)

    for i, (FP_x, FP_y) in enumerate(FP.T):
        FP_x, FP_y = FP.T[i]
        x0, y0, x1, y1 = FP_x-dx, FP_y-dy, FP_x+dx+1, FP_y+dy+1 # +1 for range

        # shuffle indicies of pixels in region
        xv, yv = [_.ravel() for _ in np.meshgrid(range(x0, x1), range(y0, y1))]
        shuffled_inds = np.random.permutation(xv.size)
        xv = xv[shuffled_inds]
        yv = yv[shuffled_inds]

        # set up new set of samples seeded with old samples (zero counts)
        n_valid = np.sum(~(samples[i] == -999)[:, 0]) # remove any with -999
        S_new = samples[i, :n_valid, :]
        C_new = np.zeros_like(counts[i][:n_valid])

        # iterate through pixels of the patch
        for x, y in zip(xv, yv):
            pixel_value = frame[y, x, :] # [R, G, B]

            # calculate distance between samples and all pixel values
            distances = np.linalg.norm(S_new - pixel_value, axis=1)

            # see which lie within R of one another
            matches = distances < R

            # if there are any matches
            if np.any(matches):
                # if there is only one, increment its counter
                if np.count_nonzero(matches) == 1:
                    C_new[matches] += 1

                # else more than one does. so increment the closest counter
                else:
                    C_new[np.argmin(distances)] += 1

            # else make new bin with count of 1
            else:
                S_new = np.append(S_new, np.array([pixel_value]), axis=0)
                C_new = np.append(C_new, np.array([1]))

        # update counts new smoothed count = (1-update_rate)*old + update_rate*new
        # so if no old count exists old = 0
        C_new *= update_rate 
        C_new[:n_valid] += (1-update_rate)*counts[i][:n_valid]

        # sort the counts by amount of pixels (highest to lowest)
        # and keep only the 'n_samples' samples with the highest counts
        sorted_inds = np.argsort(C_new)[::-1][:n_samples]

        # update the samples
        n = len(C_new)
        samples[i, :n, :] = S_new[sorted_inds, :]
        counts[i, :n] = C_new[sorted_inds]

    return samples, counts

    
def FP_placement_auto(image, bbox, dx, dy, method='S', shuffle=False, display_mask=False,
                      rho_plus=1.1, rho_minus=0.9, lbda=1e-2, tau=0.5, max_fp=999):
    # perform alpha matting to segment the object in the bbox
    mask = perform_alpha_matting(image, bbox, rho_plus=rho_plus, 
                                 rho_minus=rho_minus, lbda=lbda, tau=tau)

    # label each disconnected region of the binary mask
    mask_label = measure.label(mask) 

    # calculate properties of each labelled region - ignores 0'th label
    # such that r[0] has the label 1, r[1] -> 2 etc..
    regions = measure.regionprops(mask_label) 

    # find the region with the largest area and remove all others
    areas = [r['area'] for r in regions]
    label_max_area = np.argmax(areas) + 1
    mask_label.flat[mask_label.flat != label_max_area] = 0

    # replace original mask with only largest region left
    # note: bool maps any label > 0 to 1
    mask = np.array(mask_label, dtype='bool')
    
    min_dist = np.sqrt((2*dx+1)**2 + (2*dy+1)**2)
    
    if method == 'S':
        # create mask skeleton
        mask = morphology.thin(mask, max_iter=None) # None -> go on until skel

    if method == 'U':
        #mask = morphology.binary_erosion(mask, np.ones((dy, dx)))
        pass
    
    # randomly order all points in the mask to sample from
    yy, xx = np.nonzero(mask)
    points = np.vstack((xx, yy)).T
    
    if shuffle:
        inds = np.random.permutation(yy.size)
        points = points[inds]

    FP_INIT = np.array([points[0]], dtype='int')

    n_fp = max_fp + 1 # some arbitrarily high number
    
    while n_fp > max_fp:
        # start with first value of x, y
        FP = FP_INIT
        
        # compare next value to see if within min_dist of the FPs
        for i in range(1, points.shape[0]):
            
            # if not then add it to the list of FP locations
            if not np.any(distance.cdist(FP, np.array([points[i]])) < min_dist):
                FP = np.vstack((FP, points[i]))

        # count them to see if <= max allowed
        n_fp = FP.shape[0]
        
        # increase distance between points constraint
        min_dist += 0.5

    # transpose as we like FP[0] to be the x values, FP[0] to be y
    return FP.T
    
    
def generate_bbox_from_FP(FP, dx, dy, method=0, method_param=None):
    # FP = (2, n_patches) where FP[0, i] = [x_i, y_i], 
    # the centre the i'th fpatch
    # dx, dy = pixels left/right, up/down from centre of patch
    # method = method to use: 0
    bbox = None
    
    # axis aligned bbox expanded 20% linearly in each direction
    # method_param -> half amount to expand
    if method == 0:
        # calculate bbox
        x0, y0 = np.min(FP, axis=1) - [dx, dy] - 1
        x1, y1 = np.max(FP, axis=1) + [dx, dy] + 1
        w = x1-x0
        h = y1-y0
    
        # pad bbox slightly
        d = 10 if method_param is None else method_param
        x0 -= w/d; y0 -= h/d; x1 += w/d; y1 += h/d
        
        bbox = [x0, y0, x1, y0, x1, y1, x0, y1]
    
    # transform previous bbox based on optimal affine transform
    # method_param -> ([tx, ty, rotation, scale], affine_params)
    elif method == 1:
        bbox, affine_params = method_param
        
        # turn into shape (2, npoints), calculate mean
        bb = np.reshape(bbox, (-1, 2)).T # turn into (2, npoints)
        mn = np.mean(bb, axis=1) # mean x and y coords
        
        # add row of ones below the x and y rows for the affine transform
        bb = np.concatenate((bb, np.ones(bb.shape[1])[np.newaxis, :]), axis=0)
        
        # matrix to store affine transformation in
        A = np.eye(3)
        affine_transform(A, affine_params, mn)
        bb = (A @ bb)[:2] # remove the row of ones
        bbox = np.reshape(bb.T, (-1, )) # turn back into [x0, y0, ..., x3, y3]
    
    if bbox is not None:
        return np.array(bbox)

    raise ValueError('No invalid method specified')
    
    
def create_model(image, region, R=20, n_samples=10):
    # image     = image to work on
    # region    = [x0, y0, x1, y1] -> range of x and y to work on 
    #             range is: [x0, x1) and [y0,y1)
    # R         = vibe-like radius for a match
    # n_samples = total number of samples to use (clips to this if more)
    x0, y0, x1, y1 = region

    # shuffle indices of pixels in region
    xv, yv = [_.ravel() for _ in np.meshgrid(range(x0, x1), range(y0, y1))]
    shuffled_inds = np.random.permutation(xv.size)
    xv = xv[shuffled_inds]
    yv = yv[shuffled_inds]

    # set the first pixel as a bin (as no others exist so it matches none)
    samples = np.array([image[yv[0], xv[0], :]], dtype='float')
    counts = np.array([1], dtype='int')
    
    # iterate through the remaining pixels of the image 1->all
    for x, y in zip(xv[1:], yv[1:]):
        pixel_value = image[y, x, :] # [R, G, B]

        # calculate distance between samples and all pixel values
        distances = np.linalg.norm(samples - pixel_value, axis=1)

        # see which lie within R of one another
        matches = distances < R

        # if there are any matches
        if np.any(matches):
            # if there is only one, increment its counter
            if np.count_nonzero(matches) == 1:
                counts[matches] += 1

            # else more than one does. so increment the closest counter
            else:
                counts[np.argmin(distances)] += 1

        # else make new bin with count of 1
        else:
            samples = np.append(samples, np.array([pixel_value]), axis=0)
            counts = np.append(counts, np.array([1]))

    # sort the counts by amount of pixels (highest to lowest)
    # and keep only the 'n_samples' samples with the highest counts
    sorted_inds = np.argsort(counts)[::-1][:n_samples]

    return samples[sorted_inds], counts[sorted_inds]

    
def affine_transform(A, v, mn, sin=np.sin, cos=np.cos):
    # A - 3x3 matrix with A[2, 2] = 1, rest are overwritten
    # v - [tx, ty, rotation, scale]
    # mn - mean of points (to translate to origin from before rotation)
    s_cos_theta = v[3] * cos(v[2]) # only cos needs the scaling!
    s_sin_theta = v[3] * sin(v[2])
    
    # rotation and scale
    A[0, 0] =  s_cos_theta
    A[1, 1] =  s_cos_theta
    A[1, 0] =  s_sin_theta
    A[0, 1] = -s_sin_theta
    
    # translate to origin, perform rotation and scale, 
    # translate back to original pos + translation
    A[0, 2] = -mn[0]*s_cos_theta + mn[1]*s_sin_theta + mn[0] + v[0]
    A[1, 2] = -mn[0]*s_sin_theta - mn[1]*s_cos_theta + mn[1] + v[1]

    
def generate_and_apply_affine_transforms(FP, n_GAS, GOPT_mean, GOPT_sigma,
                                         ih, iw, dx, dy, n_LWS):
    # FP = patch centres to transform
    # n_GAS = number of transforms to sample and apply
    # GOPT_mean = mean of Gaussian
    # GOPT_sigma = covariance of Gaussian
    # ih, iw = image height, width to clip by
    # patch_size = size of patch to pad in from
    # n_LWS = local optimisation window size to pad in from
    
    n_patches = FP.shape[1]
    
    GOPT_FP_store = np.zeros((n_GAS, 2, n_patches), dtype='int')
    GOPT_A = np.eye(3, dtype='float') # so GOPT_A[2, 2] = 1 always
    GOPT_FP_prev = np.ones((3, n_patches)) # so GOPT_FP_prev[2, :] = 1 always
    
    # sample from Gaussian
    GOPT_samples = np.random.multivariate_normal(mean=GOPT_mean, 
                                                 cov=GOPT_sigma, 
                                                 size=(n_GAS,))

    # apply them to the previous timestep's FPs
    GOPT_FP_prev[:2] = FP
    GOPT_FP_prev_mean = np.mean(GOPT_FP_prev[:2], axis=1)

    for i in range(n_GAS):
        # create affine matrix as defined by the samples and FP's mean
        affine_transform(GOPT_A, GOPT_samples[i], GOPT_FP_prev_mean)

        # apply transform to FPs - [:2] as third row only needed to apply transform
        GOPT_FP_store[i] = np.rint((GOPT_A @ GOPT_FP_prev)[:2]).astype('int')

    # check which ones are valid (i.e. result in patches that lie inside the image)
    offset_x = dx+n_LWS+1 # patch width/height + local optimisation size + 1
    offset_y = dy+n_LWS+1
    valid_mask = np.any(GOPT_FP_store[:, 1, :] < offset_y, axis=1) # bottom, left
    valid_mask |= np.any(GOPT_FP_store[:, 0, :] < offset_x, axis=1) # bottom, left
    valid_mask |= np.any(GOPT_FP_store[:, 1, :] > ih-1-offset_y, axis=1) # top
    valid_mask |= np.any(GOPT_FP_store[:, 0, :] > iw-1-offset_x, axis=1) # right

    invalid_inds = np.where(valid_mask)[0]
    
    stg = len(invalid_inds) # number of Samples To Generate
    c = stg # random sample pointer 
              # (starts at stg so we generate samples IN the loop)

    # replace the invalid sets of FP locations
    for i in invalid_inds:
        while True:
            # if we've ran out of samples then generate some more!
            if c == stg:
                random_samples = np.random.multivariate_normal(mean=GOPT_mean, 
                                                               cov=GOPT_sigma, 
                                                               size=(stg,))
                c = 0
            
            # get a random sample, increase pointer
            GOPT_samples[i] = random_samples[c]
            c += 1
            
            # apply the transform and store it
            affine_transform(GOPT_A, GOPT_samples[i], GOPT_FP_prev_mean)
            GOPT_FP_store[i] = np.rint((GOPT_A @ GOPT_FP_prev)[:2]).astype('int')

            # test if its valid, repeat if not
            if not (np.any(GOPT_FP_store[i, 1, :] < offset_x)
                    or np.any(GOPT_FP_store[i, 0, :] < offset_y)
                    or np.any(GOPT_FP_store[i, 1, :] > ih-1-offset_x)
                    or np.any(GOPT_FP_store[i, 0, :] > iw-1-offset_y)):
                break
                
    return GOPT_FP_store, GOPT_samples

    
def integral_histogram_samples(I, samples, FP_G, R, dx, dy, n_LWS):
    # calculate whether each pixel in the search region matches,
    # i.e. is closer <R from, each sample. search region is defined as being
    # from: rx0, ry0: least x/y value - patch size - local opt window size - 1
    #   to: rx1, ry1: highest x/y value + patch size + local opt window size + 1
    #
    # I = image to have its pixels compared from
    # samples = (n_patches, n_samples, 3) matrix containing RGB samples
    # FP_G = (n_global_samples, 2, n_patches) matrix containing centroid 
    #        locations of the transformed FPs
    # R = distance to define closeness between pixels
    # patch_size = size of the fiducial patch
    # n_LWS = local opt window size
    #
    # returns an array: hist = ((ry1-ry0, rx1-rx0, n_patches, n_samples)) 
    # containing a 1 where a pixel matches a sample.
    # note that a pixel can only match one sample in a set of samples, 
    # i.e.sum(hist[i, j, k, :]) = 0 or 1 for all combinations of i, j, k
    # also returns rx0, ry0 - the starting points in the image of the hist,
    # used to translate FP_G into the coordinates of hist.
    
    # calculate the bounds of the image to match + int hist
    rx0 = FP_G[:, 0, :].min() - (dx + 1) - n_LWS
    ry0 = FP_G[:, 1, :].min() - (dy + 1) - n_LWS
    rx1 = FP_G[:, 0, :].max() + (dx + 1) + n_LWS
    ry1 = FP_G[:, 1, :].max() + (dy + 1) + n_LWS

    # we need this as if its stored as uint8 the subtraction below will fail.
    if np.issubdtype(samples.dtype, np.unsignedinteger):
        raise ValueError('samples must not stored as unsigned integers')
    
    # squared euclidean distance calculation
    D = samples[np.newaxis, np.newaxis, :, :, :] - I[ry0:ry1, rx0:rx1, np.newaxis, np.newaxis, :]
    D *= D
    S = np.sum(D, axis=4)
    
    # create mask of samples within R of one another and convert to ints
    mask = (S < R**2).astype('int')

    # some pixels are within R of multiple samples from the same patch,
    # so mark only the closest sample as matching
    iy, ix, iz = np.where(mask.sum(axis=3) > 1)
    a = np.argmin(S[iy, ix, iz, :], axis=1)
    mask[iy, ix, iz, :] = 0
    mask[iy, ix, iz, a] = 1

    # cumulatively sum -> integral histogram
    np.cumsum(mask, axis=0, out=mask)
    np.cumsum(mask, axis=1, out=mask)

    return mask, rx0, ry0

    
def evaluate_FPs(int_hist, counts, rx0, ry0, FP_G, dx, dy):
    # int_hist = integral histogram
    # counts = (n_patches, n_samples) hists of each sample
    # rx0, ry0 = offset from image that the int_hist was created from
    #            i.e. image[ry0, rx0] refs same point as int_hist[0, 0]
    # FP_G = (n_global_samples, 2, n_patches) matrix containing centroid 
    #        locations of the transformed FPs
    # dx, dy = amount left/right, up/down to define patch centred on a FP,
    #          i.e FP_G[i, 0, j]-dx, FP_G[i, 0, j]+dx are the indices of
    #          the x values of the i'th set of fiducial patch's j'th patch.
    
    # create references to start and end point of every patch
    # note: we need -1 as for the int hist point A needs to be [-1, -1] back
    # from its actual position (worked out on pen+paper to check ^^^)
    x0 = -rx0+FP_G[:, 0, :]-dx-1 
    y0 = -ry0+FP_G[:, 1, :]-dy-1
    x1 = -rx0+FP_G[:, 0, :]+dx   # (dont need to +1 as using direct
    y1 = -ry0+FP_G[:, 1, :]+dy   #  indexing and not ranges)

    # calculate integral histogram for respective patches
    patch_inds = np.arange(FP_G.shape[2])
    hists = int_hist[y1, x1, patch_inds, :] + int_hist[y0, x0, patch_inds, :] - int_hist[y0, x1, patch_inds, :] - int_hist[y1, x0, patch_inds, :]

    # evaluate likelihood: 1 - sqrt(1 - BC) <- BC = bhattacharyya coef
    normalisation = ((2*dx+1)*(2*dy+1))**2
    
    BC = np.sum(np.sqrt(hists * counts[np.newaxis, ...] / normalisation), axis=2)
    BC[BC>1]=1
    BC[BC<0]=0
    return 1 - np.sqrt(1 - BC)

    
def locally_optimise_FP(FP, int_hist, counts, dx, dy, 
                        rx0, ry0, n_LWS, FP_LOCALOPT_DISTS):
    # FP = (2, n_patches) Fiducial patch centroids to optimise
    # int_hist = integral histogram
    # counts = histograms of each patch
    # dx, dy = left/right, up/down offsets defining a patch
    # rx0, ry0 = x, y offsets from image for integral histogram
    # n_LWS = local optimisation window size (total width/height)
    # FP_LOCALOPT_DISTS distance between each point inside the opt window
    
    # local opt storage
    nx, ny = n_LWS//2, n_LWS//2 # length to +/- to respective dimension
    # offsets for each point to look at
    xo, yo = np.meshgrid(np.arange(-nx, nx+1), 
                         np.arange(-ny, ny+1)) 
    l_opt_store = np.zeros((2, xo.size), dtype='int')
    
    # save remaking arrays each iteration
    likelihood = np.zeros(l_opt_store.shape[1])
    mask = np.zeros_like(likelihood, dtype='bool')

    # histogram normalisation constant
    normalisation = ((2*dx+1)*(2*dy+1))**2
    
    n_patches = FP.shape[1]
    
    new_patches_likelihood = np.zeros(n_patches)
    for i in range(n_patches):
        # calculate offsets from the FP to evaluate
        l_opt_store[0] = xo.flat + FP[0, i]
        l_opt_store[1] = yo.flat + FP[1, i]

        # bounds for the patch
        x0 = -rx0 + l_opt_store[0] - dx - 1
        y0 = -ry0 + l_opt_store[1] - dy - 1
        x1 = -rx0 + l_opt_store[0] + dx
        y1 = -ry0 + l_opt_store[1] + dy

        # calc the feature vector from the integral histogram
        hists = int_hist[y1, x1, i, :] + int_hist[y0, x0, i, :] - int_hist[y0, x1, i, :] - int_hist[y1, x0, i, :]

        # bhattacharyya dist
        BC = np.sum(np.sqrt(hists * counts[i, :] / normalisation), axis=1)
        BC[BC<0]=0
        BC[BC>1]=1
        likelihood[:] = 1 - np.sqrt(1 - BC)
        am = np.argmax(likelihood)

        # check if we have a draw, if so pick the nearest one to the middle of the square
        mask[:] = likelihood == likelihood[am]
        if np.count_nonzero(mask) > 1:
            # find the indices of the FPs that are closest to the middle with the max likelihood
            am = np.where((FP_LOCALOPT_DISTS == FP_LOCALOPT_DISTS[mask].min()) & mask)[0]
            am = am[np.random.randint(am.size)] # if draw: randomly pick one
        
        new_patches_likelihood[i] = likelihood[am]
        FP[:, i] = l_opt_store[:, am]
        
    return new_patches_likelihood
    
    
def bbox_to_axis_aligned_bbox(bbox):
    """
    Converts bbox of the form [x0, y0, x1, y1, x2, y2, x3, y3] to
    axis-aligned version, i.e.: [x0, y0, width height]
    """
    xmin, xmax = np.min(bbox[::2]), np.max(bbox[::2])
    ymin, ymax = np.min(bbox[1::2]), np.max(bbox[1::2])

    return np.array([xmin, ymin, xmax-xmin, ymax-ymin])
          
          
def image2cn(image, w2c):
    # image = [..., 3], w2c = [inds, d]
    # return w2c version of image
    
    R, G, B = image[... ,0].ravel(), image[... ,1].ravel(), image[... ,2].ravel()
    inds = (np.floor(R/8) + 32*np.floor(G/8) + 32*32*np.floor(B/8)).astype('int')
    return np.reshape(w2c[inds, :], (*image.shape[:2], w2c.shape[1]))